import numpy as np
from sklearn.model_selection import train_test_split
import catboost


class CatBoostClassifier:

    def __init__(self, **params):
        """
        Catboost Classifier wrapper
        """
        default = {'verbose': 0, 'n_estimators': 1000, 'allow_writing_files': False}
        for k, v in default.items():
            if k not in params.keys():
                params[k] = v
        self.params = params
        self.model = catboost.CatBoostClassifier(**params)
        self.hasPredictProba = True
        self.classes_ = None
        self.trained = False
        self.callbacks = None
        self.verbose = 0
        self.early_stopping_rounds = 100
        if 'early_stopping_rounds' in params.keys():
            self.early_stopping_rounds = params.pop('early_stopping_rounds')
        if 'verbose' in params.keys():
            self.verbose = params.pop('verbose')
        self.set_params(**params)
        self._estimator_type = 'classifier'

    def set_params(self, **params):
        self.model.set_params(**params)
        return self

    def get_params(self, **args):
        return self.model.get_params(**args)

    def fit(self, x, y):
        # Split data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1)

        # Set Attributes
        self.classes_ = np.unique(y)

        # Train model
        self.model.fit(train_x, train_y, eval_set=[(test_x, test_y)], verbose=self.verbose,
                       early_stopping_rounds=self.early_stopping_rounds)

        # Set trained
        self.trained = True

    def predict(self, x):
        return self.model.predict(x).reshape(-1)

    def predict_proba(self, x):
        return self.model.predict_proba(x)



import re
import os
import time
import copy
import json
import joblib
import pickle
import shutil
import textwrap
import warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
from datetime import datetime

from sklearn import metrics
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler

from . import Utils

from .AutoML.Sequence import Sequence
from .AutoML.Modelling import Modelling
from .AutoML.DataExploring import DataExploring
from .AutoML.DataProcessing import DataProcessing
from .AutoML.FeatureProcessing import FeatureProcessing

from .GridSearch.BaseGridSearch import BaseGridSearch
from .GridSearch.HalvingGridSearch import HalvingGridSearch
from .GridSearch.OptunaGridSearch import OptunaGridSearch

from .Documenting.MultiDocumenting import MultiDocumenting
from .Documenting.BinaryDocumenting import BinaryDocumenting
from .Documenting.RegressionDocumenting import RegressionDocumenting


# noinspection PyUnresolvedReferences
class Pipeline:
    # todo integrate minorVersion (also with folder)
    # todo check whether normalization is beneficial

    def __init__(self,
                 target,
                 project='',
                 device='',
                 issue='',
                 version=None,
                 mode='regression',
                 objective='r2',
                 fast_run=False,

                 # Data Processing
                 num_cols=None,
                 date_cols=None,
                 cat_cols=None,
                 missing_values='interpolate',
                 outlier_removal='clip',
                 z_score_threshold=4,
                 include_output=False,

                 # Feature Processing
                 max_lags=10,
                 max_diff=2,
                 information_threshold=0.99,
                 extract_features=True,
                 feature_timeout=3600,

                 # Sequencing
                 sequence=False,
                 back=0,
                 forward=0,
                 shift=0,
                 diff='none',

                 # Initial Modelling
                 normalize=False,
                 shuffle=True,
                 cv_splits=3,
                 store_models=False,

                 # Grid Search
                 grid_search_type='optuna',
                 grid_search_time_budget=3600,
                 grid_search_candidates=250,
                 grid_search_iterations=3,

                 # Stacking
                 stacking=False,

                 # Production
                 custom_code='',

                 # Flags
                 plot_eda=True,
                 process_data=True,
                 document_results=True,
                 verbose=1):
        # Production initiation
        self.bestModel = None
        self.bestFeatures = None
        self.bestScaler = None
        self.bestOutputScaler = None

        # Starting
        print('\n\n*** Starting Amplo AutoML - {} ***\n\n'.format(project))

        # Parsing input
        self.mainDir = 'AutoML/'
        self.target = re.sub('[^a-z0-9]', '_', target.lower())
        self.verbose = verbose
        self.customCode = custom_code
        self.fastRun = fast_run
        self.project = project
        self.device = device
        self.issue = issue

        # Checks
        assert mode == 'regression' or mode == 'classification', 'Supported modes: regression, classification.'
        assert isinstance(target, str), 'Target needs to be of type string, key of target'
        assert isinstance(project, str), 'Project is a name, needs to be of type string'
        assert isinstance(num_cols, list) or num_cols is None, 'Num cols must be a list of strings'
        assert isinstance(date_cols, list) or date_cols is None, 'Date cols must be a list of strings'
        assert isinstance(cat_cols, list) or cat_cols is None, 'Cat Cols must be a list of strings'
        assert isinstance(shift, int), 'Shift needs to be an integer'
        assert isinstance(max_diff, int), 'max_diff needs to be an integer'
        assert max_lags < 50, 'Max_lags too big. Max 50.'
        assert 0 < information_threshold < 1, 'Information threshold needs to be within [0, 1'
        assert max_diff < 5, 'Max diff too big. Max 5.'
        assert isinstance(grid_search_type, str), 'Grid Search Type must be string'
        assert grid_search_type.lower() in ['base', 'halving', 'optuna'], 'Grid Search Type must be Base, Halving or ' \
                                                                          'Optuna'

        # Objective
        if objective is not None:
            self.objective = objective
        else:
            if mode == 'regression':
                self.objective = 'neg_mean_squared_error'
            elif mode == 'classification':
                self.objective = 'accuracy'
        assert isinstance(objective, str), 'Objective needs to be a string, not type {}'.format(type(objective))
        assert objective in metrics.SCORERS.keys(), 'Metric not supported, look at sklearn.metrics.SCORERS.keys()'
        self.scorer = metrics.SCORERS[self.objective]

        # Pipeline Params
        self.mode = mode
        self.version = version
        self.includeOutput = include_output
        self.plotEDA = plot_eda
        self.processData = process_data
        self.documentResults = document_results

        # Data Processing params
        self.numCols = [] if num_cols is None else num_cols
        self.dateCols = [] if date_cols is None else date_cols
        self.catCols = [] if cat_cols is None else cat_cols
        self.missingValues = missing_values
        self.outlierRemoval = outlier_removal
        self.zScoreThreshold = z_score_threshold

        # Feature Processing params
        self.extractFeatures = extract_features
        self.maxLags = max_lags
        self.maxDiff = max_diff
        self.informationThreshold = information_threshold
        self.featureTimeout = feature_timeout

        # Sequence Params
        self.sequence = sequence
        self.sequenceBack = back
        self.sequenceForward = forward
        self.sequenceShift = shift
        self.sequenceDiff = diff

        # Modelling Params
        self.normalize = normalize
        self.shuffle = shuffle
        self.cvSplits = cv_splits
        self.storeModels = store_models

        # Grid Search params
        self.gridSearchType = grid_search_type.lower()
        self.gridSearchTimeout = grid_search_time_budget
        self.gridSearchCandidates = grid_search_candidates
        self.gridSearchIterations = grid_search_iterations

        # Stacking
        self.stacking = stacking

        # Instance initiating
        self.x = None
        self.y = None
        self.colKeep = None
        self.results = None
        self.n_classes = None

        # Flags
        self._set_flags()

        # Create Directories
        self._create_dirs()

        # Load Version
        self._load_version()

        # Required sub-classes
        self.dataProcessor = DataProcessing(target=self.target, num_cols=self.numCols, date_cols=self.dateCols,
                                            cat_cols=self.catCols, missing_values=self.missingValues, mode=self.mode,
                                            outlier_removal=self.outlierRemoval, z_score_threshold=self.zScoreThreshold,
                                            folder=self.mainDir + 'Data/', version=self.version)
        self.featureProcessor = FeatureProcessing(mode=self.mode, max_lags=self.maxLags, max_diff=self.maxDiff,
                                                  extract_features=self.extractFeatures, timeout=self.featureTimeout,
                                                  information_threshold=self.informationThreshold,
                                                  folder=self.mainDir + 'Features/', version=self.version)

    def _set_flags(self):
        if self.plotEDA is None:
            self.plotEDA = Utils.boolean_input('Make all EDA graphs?')
        if self.processData is None:
            self.processData = Utils.boolean_input('Process/prepare data?')
        if self.documentResults is None:
            self.documentResults = Utils.boolean_input('Validate results?')

    def _load_version(self):
        if self.version is None:
            versions = os.listdir(self.mainDir + 'Production')
            # Updates changelog
            if self.processData:
                if len(versions) == 0:
                    if self.verbose > 0:
                        print('[AutoML] No Production files found. Setting version 0.')
                    self.version = 0
                    file = open(self.mainDir + 'changelog.txt', 'w')
                    file.write('Dataset changelog. \nv0: Initial')
                    file.close()
                else:
                    self.version = len(versions)

                    # Check if not already started
                    with open(self.mainDir + 'changelog.txt', 'r') as f:
                        changelog = f.read()

                    # Else ask for changelog
                    if 'v{}'.format(self.version) in changelog:
                        changelog = changelog[changelog.find('v{}'.format(self.version)):]
                        changelog = changelog[:max(0, changelog.find('\n'))]
                    else:
                        changelog = '\nv{}: '.format(self.version) + input("Data changelog v{}:\n".format(self.version))
                        file = open(self.mainDir + 'changelog.txt', 'a')
                        file.write(changelog)
                        file.close()
                    if self.verbose > 0:
                        print('[AutoML] Set version {}'.format(changelog[1:]))
            else:
                if len(versions) == 0:
                    if self.verbose > 0:
                        print('[AutoML] No Production files found. Starting fresh.')
                    file = open(self.mainDir + 'changelog.txt', 'w')
                    file.write('Dataset changelog. \nv0: Initial')
                    file.close()
                    self.version = 0
                else:
                    self.version = int(len(versions)) - 1
                    with open(self.mainDir + 'changelog.txt', 'r') as f:
                        changelog = f.read()
                    changelog = changelog[changelog.find('v{}'.format(self.version)):]
                    if self.verbose > 0:
                        print('[AutoML] Loading last version ({})'.format(changelog[:changelog.find('\n')]))
                    self.bestModel = joblib.load(self.mainDir + 'Production/v{}/Model.joblib'.format(self.version))
                    self.bestFeatures = json.load(open(self.mainDir + 'Production/v{}/Features.json'.format(
                        self.version), 'r'))
                    if self.normalize:
                        self.bestScaler = pickle.load(open(
                            self.mainDir + 'Production/v{}/Scaler.pickle'.format(self.version), 'rb'))
                        if self.mode == 'regression':
                            self.bestOutputScaler = pickle.load(open(
                                self.mainDir + 'Production/v{}/OutputScaler.pickle'.format(self.version), 'rb'))

    def _create_dirs(self):
        folders = ['', 'EDA', 'Data', 'Features', 'Documentation', 'Production']
        for folder in folders:
            try:
                os.makedirs(self.mainDir + folder)
            except FileExistsError:
                continue

    @staticmethod
    def _sort_results(results):
        return results.sort_values('worst_case', ascending=False)

    def _get_best_params(self, model, feature_set):
        # Filter results for model and version
        results = self.results[np.logical_and(
            self.results['model'] == type(model).__name__,
            self.results['version'] == self.version,
        )]

        # Filter results for feature set & sort them
        results = self._sort_results(results[results['dataset'] == feature_set])

        # Warning for unoptimized results
        if 'Hyper Parameter' not in results['type'].values:
            warnings.warn('Hyper parameters not optimized for this combination')

        # Parse & return best parameters (regardless of if it's optimized)
        return Utils.parse_json(results.iloc[0]['params'])

    def fit(self, data):
        """
        Fit the full autoML pipeline.
        2. (optional) Exploratory Data Analysis
        Creates a ton of plots which are helpful to improve predictions manually
        3. Data Processing
        Cleans all the data. See @DataProcessing
        4. Feature Processing
        Extracts & Selects. See @FeatureProcessing
        5. Initial Modelling
        Runs 12 off the shelf models with default parameters for all feature sets
        If Sequencing is enabled, this is where it happens, as here, the feature set is generated.
        6. Grid Search
        Optimizes the hyper parameters of the best performing models
        7. Prepare Production Files
        Nicely organises all required scripts / files to make a prediction

        @param data: DataFrame including target
        """
        # Tests
        assert isinstance(data, pd.DataFrame), 'Data must be Pandas'
        assert len(data) > 0, 'Dataframe has length zero'
        assert self.target in Utils.clean_keys(data).keys(), 'Target missing in data'

        # Execute pipeline
        self._data_processing(data)
        self._eda()
        self._feature_processing()
        self._initial_modelling()
        self.grid_search()
        self._create_stacking()
        if not os.path.exists(self.mainDir + 'Production/v{}/'.format(self.version)) or \
                len(os.listdir(self.mainDir + 'Production/v{}/'.format(self.version))) == 0:
            self._prepare_production_files()
        print('[AutoML] Done :)')

    def _eda(self):
        if self.plotEDA:
            print('[AutoML] Starting Exploratory Data Analysis')
            eda = DataExploring(self.x, y=self.y,
                                mode=self.mode,
                                folder=self.mainDir,
                                version=self.version)
            eda.run()

    def _data_processing(self, data):
        # Load if possible
        if os.path.exists(self.mainDir + 'Data/Cleaned_v{}.csv'.format(self.version)):
            print('[AutoML] Loading Cleaned Data')
            data = pd.read_csv(self.mainDir + 'Data/Cleaned_v{}.csv'.format(self.version), index_col='index')

        # Clean
        else:
            print('[AutoML] Cleaning Data')
            data = self.dataProcessor.clean(data)

        # Split and store in memory
        self.y = data[self.target]
        self.x = data
        if self.includeOutput is False:
            self.x = self.x.drop(self.target, axis=1)

        # Assert classes in case of classification
        self.n_classes = self.y.nunique()
        if self.mode == 'classification':
            if self.n_classes >= 50:
                warnings.warn('More than 50 classes, you may want to reconsider classification mode')
            if set(self.y) != set([i for i in range(len(set(self.y)))]):
                raise ValueError('Classes should be [0, 1, ...]')

    def _feature_processing(self):
        # Check if exists
        if os.path.exists(self.mainDir + 'Data/Extracted_v{}.csv'.format(self.version)):
            print('[AutoML] Loading Extracted Features')
            self.x = pd.read_csv(self.mainDir + 'Data/Extracted_v{}.csv'.format(self.version), index_col='index')
            self.colKeep = json.load(open(self.mainDir + 'Features/Sets_v{}.json'.format(self.version), 'r'))

        else:
            # Extract
            self.x = self.featureProcessor.extract(self.x, self.y)

            # Select
            self.colKeep = self.featureProcessor.select(self.x, self.y)

            # Store
            self.x.to_csv(self.mainDir + 'Data/Extracted_v{}.csv'.format(self.version), index_label='index')

    def _initial_modelling(self):
        # Load existing results
        if 'Results.csv' in os.listdir(self.mainDir):
            self.results = pd.read_csv(self.mainDir + 'Results.csv')

            # Printing here as we load it
            results = self.results[np.logical_and(
                self.results['version'] == self.version,
                self.results['type'] == 'Initial modelling'
            )]
            for fs in set(results['dataset']):
                print('[AutoML] Initial Modelling for {} ({})'.format(fs, len(self.colKeep[fs])))
                fsr = results[results['dataset'] == fs]
                for i in range(len(fsr)):
                    row = fsr.iloc[i]
                    print('[AutoML] {} {}: {:.4f} \u00B1 {:.4f}'.format(row['model'].ljust(40), self.objective,
                                                                        row['mean_objective'], row['std_objective']))

        # Check if this version has been modelled
        if self.results is not None and self.version in self.results['version'].values:
            self.results = self._sort_results(self.results)

        # Run Modelling
        else:
            for feature_set, cols in self.colKeep.items():
                # Skip empty sets
                if len(cols) == 0:
                    print('[AutoML] Skipping {} features, empty set'.format(feature_set))
                else:
                    print('[AutoML] Initial Modelling for {} features ({})'.format(feature_set, len(cols)))

                    # Select data
                    x, y = copy.deepcopy(self.x[cols]), copy.deepcopy(self.y)

                    # Normalize Feature Set (Done here to get one normalization file per feature set)
                    if self.normalize:
                        normalize_features = [k for k in x.keys() if k not in self.dateCols + self.catCols]
                        scaler = StandardScaler()
                        x[normalize_features] = scaler.fit_transform(x[normalize_features])
                        pickle.dump(scaler,
                                    open(self.mainDir + 'Features/Scaler_{}_{}.pickle'.format(
                                        feature_set, self.version), 'wb'))
                        if self.mode == 'regression':
                            output_scaler = StandardScaler()
                            y = pd.Series(output_scaler.fit_transform(y.values.reshape(-1, 1)).reshape(-1),
                                          name=self.target)
                            pickle.dump(output_scaler,
                                        open(self.mainDir + 'Features/OutputScaler_{}_{}.pickle'.format(
                                            feature_set, self.version),
                                             'wb'))

                    # Sequence if necessary
                    if self.sequence:
                        sequencer = Sequence(back=self.sequenceBack, forward=self.sequenceForward,
                                             shift=self.sequenceShift, diff=self.sequenceDiff)
                        x, y = sequencer.convert(x, y)

                    # Do the modelling
                    modeller = Modelling(mode=self.mode, shuffle=self.shuffle, store_models=self.storeModels,
                                         objective=self.objective, dataset=feature_set,
                                         store_results=False, folder=self.mainDir + 'Models/')
                    results = modeller.fit(x, y)

                    # Add results to memory
                    results['type'] = 'Initial modelling'
                    results['version'] = self.version
                    if self.results is None:
                        self.results = results
                    else:
                        self.results = self.results.append(results)

            # Save results
            self.results.to_csv(self.mainDir + 'Results.csv', index=False)

    def grid_search(self, model=None, feature_set=None, parameter_set=None):
        """
        Runs a grid search. By default, takes the self.results, and runs for the top 3 optimizations.
        There is the option to provide a model & feature_set, but both have to be provided. In this case,
        the model & data set combination will be optimized.
        Implemented types, Base, Halving, Optuna
        """
        assert model is not None and feature_set is not None or model == feature_set, \
            'Model & feature_set need to be either both None or both provided.'
        # If arguments are provided
        if model is not None:

            # Get model string
            if isinstance(model, str):
                models = Modelling(mode=self.mode, samples=len(self.x), objective=self.objective).return_models()
                model = models[[i for i in range(len(models)) if type(models[i]).__name__ == model][0]]

            # Organise existing results
            results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['version'] == self.version,
            )]
            results = self._sort_results(results[results['dataset'] == feature_set])

            # Check if exists and load
            if ('Hyper Parameter' == results['type']).any():
                print('[AutoML] Loading optimization results.')
                hyper_opt_results = results[results['type'] == 'Hyper Parameter']
                params = Utils.parse_json(hyper_opt_results.iloc[0]['params'])

            # Or run
            else:
                # Run grid search
                grid_search_results = self._sort_results(self._grid_search_iteration(model, parameter_set, feature_set))

                # Store results
                grid_search_results['model'] = type(model).__name__
                grid_search_results['version'] = self.version
                grid_search_results['dataset'] = feature_set
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)

                # Get params for validation
                params = Utils.parse_json(grid_search_results.iloc[0]['params'])

            # Validate
            if self.documentResults:
                self.document(model.set_params(**params), feature_set)
            return

        # If arguments aren't provided, run through promising models
        models = Modelling(mode=self.mode, samples=len(self.x), objective=self.objective).return_models()
        results = self._sort_results(self.results[np.logical_and(
            self.results['type'] == 'Initial modelling',
            self.results['version'] == self.version,
        )])
        for iteration in range(self.gridSearchIterations):
            # Grab settings
            settings = results.iloc[iteration]
            model = copy.deepcopy(models[[i for i in range(len(models)) if type(models[i]).__name__ ==
                                          settings['model']][0]])
            feature_set = settings['dataset']

            # Check whether exists
            model_results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['version'] == self.version,
            )]
            model_results = self._sort_results(model_results[model_results['dataset'] == feature_set])

            # If exists
            if ('Hyper Parameter' == model_results['type']).any():
                hyper_opt_res = model_results[model_results['type'] == 'Hyper Parameter']
                params = Utils.parse_json(hyper_opt_res.iloc[0]['params'])

            # Else run
            else:
                # For one model
                grid_search_results = self._sort_results(self._grid_search_iteration(
                    copy.deepcopy(model), parameter_set, feature_set))

                # Store
                grid_search_results['version'] = self.version
                grid_search_results['dataset'] = feature_set
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)
                params = Utils.parse_json(grid_search_results.iloc[0]['params'])

            # Validate
            if self.documentResults:
                self.document(model.set_params(**params), feature_set)

    def _grid_search_iteration(self, model, parameter_set, feature_set):
        """
        INTERNAL | Grid search for defined model, parameter set and feature set.
        """
        print('\n[AutoML] Starting Hyper Parameter Optimization for {} on {} features ({} samples, {} features)'.format(
            type(model).__name__, feature_set, len(self.x), len(self.colKeep[feature_set])))

        # Select data
        x, y = copy.deepcopy(self.x[self.colKeep[feature_set]]), copy.deepcopy(self.y)

        # Normalize Feature Set (the input remains original)
        if self.normalize:
            features_to_normalize = [k for k in self.colKeep[feature_set] if k not in self.dateCols + self.catCols]
            scaler = pickle.load(
                open(self.mainDir + 'Features/Scaler_{}_{}.pickle'.format(feature_set, self.version), 'rb'))
            x[features_to_normalize] = scaler.transform(x[features_to_normalize])
            if self.mode == 'regression':
                output_scaler = pickle.load(
                    open(self.mainDir + 'Features/OutputScaler_{}_{}.pickle'.format(feature_set, self.version), 'rb'))
                y = pd.Series(output_scaler.transform(y.values.reshape(-1, 1)).reshape(-1), name=self.target)

        # Cross-Validator
        cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)
        if self.mode == 'regression':
            cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)

        # Select right hyper parameter optimizer
        if self.gridSearchType == 'base':
            grid_search = BaseGridSearch(model, params=parameter_set, cv=cv, scoring=self.objective,
                                         candidates=self.gridSearchCandidates, timeout=self.gridSearchTimeout,
                                         verbose=self.verbose)
        elif self.gridSearchType == 'halving':
            grid_search = HalvingGridSearch(model, params=parameter_set, cv=cv, scoring=self.objective,
                                            candidates=self.candidates, verbose=self.verbose)
        elif self.gridSearchType == 'optuna':
            grid_search = OptunaGridSearch(model, params=parameter_set, timeout=self.gridSearchTimeout, cv=cv,
                                           candidates=self.gridSearchCandidates, scoring=self.objective,
                                           verbose=self.verbose)
        else:
            raise NotImplementedError('Only Base, Halving and Optuna are implemented.')
        # Get results
        results = grid_search.fit(x, y)
        return results.sort_values('worst_case', ascending=False)

    def _create_stacking(self):
        """
        Based on the best performing models, in addition to cheap models based on very different assumptions,
        A stacking model is optimized to enhance/combine the performance of the models.
        --> should contain a large variety of models
        --> classifiers need predict_proba
        --> level 1 needs to be ordinary least squares
        """
        if self.stacking:
            print('[AutoML] Creating Stacking Ensemble')
            from sklearn import neighbors
            from sklearn import tree
            from sklearn import linear_model
            from sklearn import svm
            from sklearn import naive_bayes
            from sklearn import ensemble

            # Select feature set that has been picked most often for hyper parameter optimization
            results = self._sort_results(self.results[np.logical_and(
                self.results['type'] == 'Hyper Parameter',
                self.results['version'] == self.version,
            )])
            feature_set = results['dataset'].value_counts().index[0]
            print('[AutoML] Selected Stacking feature set: {}'.format(feature_set))
            results = results[results['dataset'] == feature_set]

            # Level 0, top n models + KNN, DT, Log Reg, (GNB), (SVM)
            n_stacking_models = 3
            stacking_models_str = results['model'].unique()[:n_stacking_models]
            stacking_models_params = [Utils.parse_json(results.iloc[np.where(results['model'] == sms)[0][0]]['params'])
                                      for sms in stacking_models_str]
            models = Modelling(mode=self.mode, samples=len(self.x), objective=self.objective).return_models()
            models_str = [type(m).__name__ for m in models]
            stacking_models = [(sms, models[models_str.index(sms)].set_params(**stacking_models_params[i]))
                               for i, sms in enumerate(stacking_models_str)]

            # Prepare Stack
            if self.mode == 'regression':
                if 'KNeighborsRegressor' not in stacking_models_str:
                    stacking_models.append(('KNeighborsRegressor', neighbors.KNeighborsRegressor()))
                if 'DecisionTreeRegressor' not in stacking_models_str:
                    stacking_models.append(('DecisionTreeRegressor', tree.DecisionTreeRegressor()))
                if 'LinearRegression' not in stacking_models_str:
                    stacking_models.append(('LinearRegression', linear_model.LinearRegression()))
                if 'SVR' not in stacking_models_str and len(self.x) < 5000:
                    stacking_models.append(('SVR', svm.SVR()))
                level_one = linear_model.LinearRegression()
                stack = ensemble.StackingRegressor(stacking_models, final_estimator=level_one)
                cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            elif self.mode == 'classification':
                if 'KNeighborsClassifier' not in stacking_models_str:
                    stacking_models.append(('KNeighborsClassifier', neighbors.KNeighborsClassifier()))
                if 'DecisionTreeClassifier' not in stacking_models_str:
                    stacking_models.append(('DecisionTreeClassifier', tree.DecisionTreeClassifier()))
                if 'LogisticRegression' not in stacking_models_str:
                    stacking_models.append(('LogisticRegression', linear_model.LogisticRegression()))
                if 'GaussianNB' not in stacking_models_str:
                    stacking_models.append(('GaussianNB', naive_bayes.GaussianNB()))
                if 'SVC' not in stacking_models_str and len(self.x) < 5000:
                    stacking_models.append(('SVC', svm.SVC()))
                level_one = linear_model.LogisticRegression(max_iter=1000)
                stack = ensemble.StackingClassifier(stacking_models, final_estimator=level_one)
                cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            else:
                raise NotImplementedError('Unknown mode')
            print('[AutoML] Stacked models: {}'.format([type(i[1]).__name__ for i in stacking_models]))

            # Train stack
            x, y = copy.deepcopy(self.x[self.colKeep[feature_set]]), copy.deepcopy(self.y)
            if self.normalize:
                normalize_features = [k for k in x.keys() if k not in self.dateCols + self.catCols]
                scaler = pickle.load(open(
                    self.mainDir + 'Features/Scaler_{}_{}.pickle'.format(feature_set, self.version), 'rb'))
                x[normalize_features] = scaler.fit_transform(x[normalize_features])
                if self.mode == 'regression':
                    output_scaler = pickle.load(open(self.mainDir + 'Features/OutputScaler_{}_{}.pickle'.format(
                        feature_set, self.version), 'rb'))
                    y = pd.Series(output_scaler.fit_transform(y.values.reshape(-1, 1)).reshape(-1),
                                  name=self.target)
            # Sequence if necessary
            if self.sequence:
                sequencer = Sequence(back=self.sequenceBack, forward=self.sequenceForward,
                                     shift=self.sequenceShift, diff=self.sequenceDiff)
                x, y = sequencer.convert(x, y)
            if isinstance(x, pd.DataFrame):
                x, y = x.to_numpy(), y.to_numpy()

            # Cross Validate
            score = []
            times = []
            for (t, v) in tqdm(cv.split(x, y)):
                start_time = time.time()
                xt, xv, yt, yv = x[t], x[v], y[t].reshape((-1)), y[v].reshape((-1))
                model = copy.deepcopy(stack)
                model.fit(xt, yt)
                score.append(self.scorer(model, xv, yv))
                times.append(time.time() - start_time)
            print('[AutoML] Stacking result:')
            print('[AutoML] {}:        {:.2f} \u00B1 {:.2f}'.format(self.objective, np.mean(score), np.std(score)))
            self.results.append(pd.DataFrame({
                'date': datetime.today().strftime('%d %b %y'),
                'model': type(stack).__name__,
                'dataset': feature_set,
                'params': dict([(stacking_models_str[ind] + '__' + key, stacking_models_params[ind][key]) for ind in
                                range(len(stacking_models_str)) for key in stacking_models_params[ind].keys()]),
                'mean_objective': np.mean(score),
                'std_objective': np.std(score),
                'mean_time': np.mean(times),
                'std_time': np.std(times),
                'version': self.version,
                'type': 'Stacking',
            }))
            self.results.to_csv(self.mainDir + 'Results.csv', index=False)
            if self.documentResults:
                self.document(stack, feature_set)

    def document(self, model, feature_set):
        """
        Loads the model and features and initiates the outside Documenting class.
        """
        # Get model
        if isinstance(model, str):
            models = Modelling(mode=self.mode, samples=len(self.x), objective=self.objective).return_models()
            model = models[[i for i in range(len(models)) if type(models[i]).__name__ == model][0]]

        assert feature_set in self.colKeep.keys(), 'Feature Set not available.'
        if os.path.exists(self.mainDir + 'Documentation/v{}/{}_{}.pdf'.format(
                self.version, type(model).__name__, feature_set)):
            print('[AutoML] Documentation existing for {} v{} - {} '.format(
                type(model).__name__, self.version, feature_set))
            return
        if len(model.get_params()) == 0:
            warnings.warn('[Documenting] Supplied model has no parameters!')

        # Run validation
        print('[AutoML] Creating Documentation for {} - {}'.format(type(model).__name__, feature_set))
        if self.mode == 'classification' and self.n_classes == 2:
            documenting = BinaryDocumenting(self)
        elif self.mode == 'classification':
            documenting = MultiDocumenting(self)
        elif self.mode == 'regression':
            documenting = RegressionDocumenting(self)
        documenting.create(model, feature_set)

    def _prepare_production_files(self, model=None, feature_set=None, params=None):
        if not os.path.exists(self.mainDir + 'Production/v{}/'.format(self.version)):
            os.mkdir(self.mainDir + 'Production/v{}/'.format(self.version))
        # Get sorted results for this data version
        results = self._sort_results(self.results[self.results['version'] == self.version])

        # In the case args are provided
        if model is not None and feature_set is not None:
            # Take name if model instance is given
            if not isinstance(model, str):
                model = type(model).__name__
            if params is None:
                results = self._sort_results(
                    results[np.logical_and(results['model'] == model, results['dataset'] == feature_set)])
                params = Utils.parse_json(results.iloc[0]['params'])

        # Otherwise find best
        else:
            model = results.iloc[0]['model']
            feature_set = results.iloc[0]['dataset']
            params = results.iloc[0]['params']
            if isinstance(params, str):
                params = Utils.parse_json(params)

        # Stacking Warning
        if 'Stacking' in model:
            warnings.warn('Stacking Models not Production Ready, skipping to next best')
            model = results.iloc[1]['model']
            feature_set = results.iloc[1]['dataset']
            params = Utils.parse_json(results.iloc[1]['params'])

        # Save Features
        self.bestFeatures = self.colKeep[feature_set]
        json.dump(self.bestFeatures, open(self.mainDir + 'Production/v{}/Features.json'.format(self.version), 'w'))

        # Copy data
        x, y = copy.deepcopy(self.x[self.bestFeatures]), copy.deepcopy(self.y)

        # Save Scaler & Normalize
        if self.normalize:
            # Save
            shutil.copy(self.mainDir + 'Features/Scaler_{}_{}.pickle'.format(feature_set, self.version),
                        self.mainDir + 'Production/v{}/Scaler.pickle'.format(self.version))
            if self.mode == 'regression':
                shutil.copy(self.mainDir + 'Features/OutputScaler_{}_{}.pickle'.format(feature_set, self.version),
                            self.mainDir + 'Production/v{}/OutputScaler.pickle'.format(self.version))

            # Normalize
            normalize_features = [k for k in self.bestFeatures if k not in self.dateCols + self.catCols]
            self.bestScaler = pickle.load(
                open(self.mainDir + 'Features/Scaler_{}_{}.pickle'.format(feature_set, self.version), 'rb'))
            x = self.bestScaler.transform(x[normalize_features])
            if self.mode == 'regression':
                self.bestOutputScaler = pickle.load(
                    open(self.mainDir + 'Features/OutputScaler_{}_{}.pickle'.format(feature_set, self.version), 'rb'))
                y = self.bestOutputScaler.transform(y.values.reshape(-1, 1)).reshape(-1)
            else:
                y = self.y.to_numpy()

        # Cluster Features require additional 'Centers' file
        if any(['dist__' in key for key in self.bestFeatures]):
            shutil.copy(self.mainDir + 'Features/KMeans_v{}.csv'.format(self.version),
                        self.mainDir + 'Production/v{}/KMeans.csv'.format(self.version))

        # Model
        models = Modelling(mode=self.mode, samples=len(self.x), objective=self.objective).return_models()
        self.bestModel = [mod for mod in models if type(mod).__name__ == model][0]
        self.bestModel.set_params(**params)
        self.bestModel.fit(x, y)
        joblib.dump(self.bestModel, self.mainDir + 'Production/v{}/Model.joblib'.format(self.version))

        # Predict function
        predict_code = self.create_predict_function(self.customCode)
        with open(self.mainDir + 'Production/v{}/Predict.py'.format(self.version), 'w') as f:
            f.write(predict_code)
        with open(self.mainDir + 'Production/v{}/__init__.py'.format(self.version), 'w') as f:
            f.write('')

        # Pipeline
        pickle.dump(self, open(self.mainDir + 'Production/v{}/Pipeline.pickle'.format(self.version), 'wb'))

        # Rapport
        if not os.path.exists('{}Documentation/v{}/{}_{}.pdf'.format(self.mainDir, self.version, model, feature_set)):
            self.document(self.bestModel, feature_set)
        shutil.copy('{}Documentation/v{}/{}_{}.pdf'.format(self.mainDir, self.version, model, feature_set),
                    '{}Production/v{}/{}_{}_v{}.pdf'.format(self.mainDir, self.version, self.device, self.issue,
                                                            self.version))

        # Notify of results
        print('[AutoML] Preparing Production Env Files for {}, feature set {}'.format(model, feature_set))
        print('[AutoML] ', params)
        print('[AutoML] Model fully fitted.')
        print('[AutoML] In-sample {}: {:4f}'.format(self.objective, self.scorer(self.bestModel, x, y)))
        return

    def _error_analysis(self):
        # todo implement
        pass

    def convert_data(self, data):
        # Load files
        folder = 'Production/v{}/'.format(self.version)
        features = self.bestFeatures

        # Clean data
        data = Utils.clean_keys(data)
        data = self.dataProcessor.convert_data_types(data)
        data = self.dataProcessor.remove_duplicates(data)
        data = self.dataProcessor.remove_outliers(data)
        data = self.dataProcessor.remove_missing_values(data)
        if data.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
            raise ValueError('Data should not contain NaN or Infinities after cleaning!')

        # Save output
        if self.target in data.keys():
            y = data[self.target].to_numpy().reshape((-1, 1))
        else:
            y = None

        # Convert Features
        if 'KMeans.csv' in os.listdir(self.mainDir + folder):
            k_means = pd.read_csv(self.mainDir + folder + 'KMeans.csv')
            x = self.featureProcessor.transform(data=data, features=features, k_means=k_means)
        else:
            x = self.featureProcessor.transform(data, features)

        if x.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
            raise ValueError('Data should not contain NaN or Infinity after adding features!')

        # Normalize
        if self.normalize:
            x[x.keys()] = self.bestScaler.transform(x)
            if self.mode == 'regression' and y is not None:
                y = self.bestOutputScaler.transform(y)

        # Return
        return x, y

    def predict(self, data):
        """
        Full script to make predictions. Uses 'Production' folder with defined or latest version.
        @param data: data to do prediction on
        """
        # Feature Extraction, Selection and Normalization
        if self.verbose > 0:
            print('[AutoML] Predicting with {}, v{}'.format(type(self.bestModel).__name__, self.version))
        x, y = self.convert_data(data)

        # Predict
        if self.mode == 'regression':
            if self.normalize:
                predictions = self.bestOutputScaler.invers_transform(self.bestModel.predict(x))
            else:
                predictions = self.bestModel.predict(x)
        elif self.mode == 'classification':
            predictions = self.bestModel.predict(x)
        else:
            raise ValueError('Unsupported mode')

        return predictions

    def predict_proba(self, data):
        """
        Returns probabilistic prediction, only for classification.
        @param data: data to do prediction on
        """
        # Tests
        assert self.mode == 'classification', 'Predict_proba only available for classification'
        assert hasattr(self.bestModel, 'predict_proba'), '{} has no attribute predict_proba'.format(
            type(self.bestModel).__name__)

        # Print
        if self.verbose > 0:
            print('[AutoML] Predicting with {}, v{}'.format(type(self.bestModel).__name__, self.version))

        # Convert data
        x, y = self.convert_data(data)

        # Predict
        return self.bestModel.predict_proba(x)

    def create_predict_function(self, custom_code):
        """
        This function returns a string, which can be used to make predictions.
        This is in a predefined format, a Predict class, with a predict function taking the arguments
        model: trained sklearn-like class with the .fit() function
        features: list of strings containing all features fed to the model
        scaler: trained sklearn-like class with .transform function
        data: the data to predict on
        Now this function has the arg decoding, which allows custom code injection
        """
        # Check if predict file exists already to increment version
        if os.path.exists(self.mainDir + 'Production/v{}/Predict.py'.format(self.version)):
            with open(self.mainDir + 'Production/v{}/Predict.py'.format(self.version), 'r') as f:
                predict_file = f.read()
            ind = predict_file.find('self.version = ') + 16
            old_version = predict_file[ind: predict_file.find("'", ind)]
            minor_version = int(old_version[old_version.find('.') + 1:])
            version = old_version[:old_version.find('.') + 1] + str(minor_version + 1)
        else:
            version = 'v{}.0'.format(self.version)
        print('Creating Prediction {}'.format(version))
        data_process = self.dataProcessor.export_function()
        feature_process = self.featureProcessor.export_function()
        return '''import re
import copy
import warnings
import itertools
import numpy as np
import pandas as pd


class Predict(object):

    def __init__(self):
        self.version = '{}'

    @staticmethod
    def predict(model, features, data, **args):
        """
        Prediction function for Amplo AutoML.
        This is in a predefined format:
        - a 'Predict' class, with a 'predict' function taking the arguments:
            model: trained sklearn-like class with the .fit() function
            features: list of strings containing all features fed to the model
            data: the data to predict on
        Note: May depend on additional named arguments within args.
        """
        ###############
        # Custom Code #
        ###############'''.format(version) + textwrap.indent(custom_code, '        ') \
               + data_process + feature_process + '''
        ###########
        # Predict #
        ###########
        mode, normalize = '{}', {}

        # Normalize
        if normalize:
            assert 'scaler' in args.keys(), 'When Normalizing=True, scaler needs to be provided in args'
            x = args['scaler'].transform(x)

        # Predict
        if mode == 'regression':
            if normalize:
                assert 'o_scaler' in args.keys(), 'When Normalizing=True, o_scaler needs to be provided in args'
                return args['OutputScaler'].inverse_transform(model.predict(x))
            else:
                return model.predict(x)
        elif mode == 'classification':
            try:
                return model.predict_proba(x)
            except AttributeError:
                return model.predict(x)
        else:
            raise NotImplementedError('Mode not supported, pick between classification and regression.')
'''.format(self.mode, self.normalize)



import os
import copy
import ppscore
import numpy as np
import pandas as pd
from tqdm import tqdm
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
from statsmodels.tsa.seasonal import STL
from statsmodels.graphics.tsaplots import plot_acf
from statsmodels.graphics.tsaplots import plot_pacf
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import RandomForestClassifier


class DataExploring:

    def __init__(self, data,
                 y=None,
                 mode='regression',
                 plot_time_plots=True,
                 plot_box_plots=False,
                 plot_missing_values=True,
                 plot_seasonality=False,
                 plot_co_linearity=True,
                 plot_differencing=False,
                 plot_signal_correlations=False,
                 plot_feature_importance=True,
                 plot_scatter_plots=False,
                 differ=0,
                 pre_tag='',
                 max_samples=10000,
                 season_periods=24,
                 lags=60,
                 skip_completed=True,
                 folder='',
                 version=0):
        """
        Automated Exploratory Data Analysis. Covers binary classification and regression.
        It generates:
        - Missing Values Plot
        - Line Plots of all features
        - Box plots of all features
        - Co-linearity Plot
        - SHAP Values
        - Random Forest Feature Importance
        - Predictive Power Score

        Additionally fFor Regression:
        - Seasonality Plots
        - Differentiated Variance Plot
        - Auto Correlation Function Plot
        - Partial Auto Correlation Function Plot
        - Cross Correlation Function Plot
        - Scatter Plots

        All figures are organized per project, category and version with the following file structure:
        |-- 'Folder parameter'
        |   |-- EDA
        |   |   |-- Timeplots
        |   |   |   |-- v'i'
        |   |   |   |   |-- Feature_0.png
        |   |   |   |   |-- Feature_1.png
        ...

        Parameters
        ----------
        data pd.DataFrame: Input data, either Numpy or Pandas
        y pd.Series: Output / Target, either Numpy or Pandas
        mode str: classification or regression
        plot_time_plots bool: Boolean to create time plots or not
        plot_box_plots bool: Boolean to create box plots
        plot_missing_values bool: Boolean to create missing values plot
        plot_seasonality bool: Boolean to create seasonality plots
        plot_co_linearity bool: Boolean to create co-linearity plots
        plot_differencing bool: Boolean to create differencing plots
        plot_signal_correlations bool: Boolean to create signal correlation plots (ACF, PACF, CCF)
        plot_feature_importance bool: Boolean to create feature importance plots (SHAP, RF, PPS)
        plot_scatter_plots bool: Boolean to create scatter plots
        differ int: Differencing order for signal correlation plots
        pre_tag str: Optional, string to format file names
        max_samples int: Optional, under-sampling for large datasets
        season_periods int: Optional, season periods for seasonality plots
        lags int: Optional, maximum lags to take into account for signal plots
        skip_completed bool: Whether to create existing plots
        folder str: Main folder to dump files in
        version int: All graphs are put in a versioned folder
        """
        assert isinstance(data, np.ndarray) or isinstance(data, pd.DataFrame), 'Data should be Numpy or Pandas'
        if y is not None:
            assert isinstance(y, np.ndarray) or isinstance(y, pd.Series), 'Y should be Numpy or Pandas'
        if type(data) != pd.DataFrame:
            data = pd.DataFrame(data=data, columns=['Feature_{}'.format(i) for i in range(data.shape[1])])
        assert mode in ['classification', 'regression']
        assert isinstance(plot_time_plots, bool)
        assert isinstance(plot_box_plots, bool)
        assert isinstance(plot_missing_values, bool)
        assert isinstance(plot_co_linearity, bool)
        assert isinstance(plot_differencing, bool)
        assert isinstance(plot_signal_correlations, bool)
        assert isinstance(plot_feature_importance, bool)
        assert isinstance(plot_scatter_plots, bool)
        assert isinstance(differ, int)
        assert isinstance(pre_tag, str)
        assert isinstance(max_samples, int)
        assert isinstance(season_periods, int)
        assert isinstance(lags, int)
        assert isinstance(skip_completed, bool)
        assert isinstance(folder, str)
        assert isinstance(version, int)

        # Running booleans
        self.plotTimePlots = plot_time_plots
        self.plotBoxPlots = plot_box_plots
        self.plotMissingValues = plot_missing_values
        self.plotSeasonality = plot_seasonality
        self.plotCoLinearity = plot_co_linearity
        self.plotDifferencing = plot_differencing
        self.plotSignalCorrelations = plot_signal_correlations
        self.plotFeatureImportance = plot_feature_importance
        self.plotScatterPlots = plot_scatter_plots

        # Register data
        self.data = data.astype('float32').fillna(0)
        if y is not None:
            if type(y) == pd.DataFrame:
                y = y[y.keys()[0]]
            if type(y) != pd.Series:
                y = pd.Series(y)
            self.Y = y.astype('float32').fillna(0)

        # General settings
        self.seasonPeriods = season_periods
        self.maxSamples = max_samples       # Time series
        self.differ = differ                # Correlations
        self.lags = lags                    # Correlations

        # Storage settings
        self.mode = mode
        self.tag = pre_tag
        self.version = version if version is not None else 0
        self.skip = skip_completed
        self.folder = folder if folder == '' or folder[-1] == '/' else folder + '/'
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

        # Create Base folder
        if not os.path.exists(self.folder + 'EDA/'):
            os.mkdir(self.folder + 'EDA')
        self.folder += 'EDA/'

    def run(self):
        # Run all functions
        print('[EDA] Generating Missing Values Plot')
        self.missing_values()
        print('[EDA] Generating Time plots')
        self.time_plots()
        print('[EDA] Generating Box plots')
        self.box_plots()
        print('[EDA] Generating Co-linearity Plots')
        self.co_linearity()

        # Functions which require output
        if self.Y is not None:
            print('[EDA] Generating SHAP plot')
            self.shap()
            print('[EDA] Generating Feature Ranking Plot')
            self.feature_ranking()
            print('[EDA] Predictive Power Score Plot')
            self.predictive_power_score()

        # Mode specific functions
        if self.mode == 'classification':
            self._run_classification()
        else:
            self._run_regression()

    def _run_classification(self):
        pass

    def _run_regression(self):
        if self.plotSeasonality:
            print('[EDA] Generating Seasonality Plots')
            self.seasonality()
        print('[EDA] Generating Diff Var Plot')
        self.differencing()
        print('[EDA] Generating ACF Plots')
        self.complete_auto_corr()
        print('[EDA] Generating PACF Plots')
        self.partial_auto_corr()
        if self.Y is not None:
            print('[EDA] Generating CCF Plots')
            self.cross_corr()
            print('[EDA] Generating Scatter plots')
            self.scatters()

    def missing_values(self):
        if self.plotMissingValues:
            # Create folder
            if not os.path.exists(self.folder + 'MissingValues/'):
                os.mkdir(self.folder + 'MissingValues/')

            # Skip if exists
            if self.tag + 'v{}.png'.format(self.version) in os.listdir(self.folder + 'MissingValues/'):
                return

            # Plot
            ax = missingno.matrix(self.data, figsize=[24, 16])
            fig = ax.get_figure()
            fig.savefig(self.folder + 'MissingValues/v{}.png'.format(self.version))

    def box_plots(self):
        if self.plotBoxPlots:
            # Create folder
            if not os.path.exists(self.folder + 'BoxPlots/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'BoxPlots/v{}/'.format(self.version))

            # Iterate through vars
            for key in tqdm(self.data.keys()):

                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'BoxPlots/v{}/'.format(self.version)):
                    continue

                # Figure prep
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Classification
                if self.mode == 'classification':
                    plt.boxplot([self.data.loc[self.Y == 1, key], self.data.loc[self.Y == -1, key]], labels=['Faulty',
                                                                                                             'Healthy'])
                    plt.legend(['Faulty', 'Healthy'])

                # Regression
                if self.mode == 'regression':
                    plt.boxplot(self.data[key])

                # Store & Close
                fig.savefig(self.folder + 'BoxPlots/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=300)
                plt.close()

    def time_plots(self):
        if self.plotTimePlots:
            # Create folder
            if not os.path.exists(self.folder + 'TimePlots/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'TimePlots/v{}/'.format(self.version))

            # Set matplot limit
            matplotlib.use('Agg')
            matplotlib.rcParams['agg.path.chunksize'] = 200000

            # Undersample
            ind = np.linspace(0, len(self.data) - 1, self.maxSamples).astype('int')
            data, y = self.data.iloc[ind], self.Y.iloc[ind]

            # Iterate through features
            for key in tqdm(data.keys()):
                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'TimePlots/v{}/'.format(self.version)):
                    continue

                # Figure preparation
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Plot
                if self.mode == 'classification':
                    cm = plt.get_cmap('bwr')
                else:
                    cm = plt.get_cmap('summer')
                nm_output = (y - y.min()) / (y.max() - y.min())
                plt.scatter(data.index, data[key], c=cm(nm_output), alpha=0.3)

                # Store & Close
                fig.savefig(self.folder + 'TimePlots/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def seasonality(self):
        # Create folder
        if not os.path.exists(self.folder + 'Seasonality/'):
            os.mkdir(self.folder + 'Seasonality/')

        # Iterate through features
        for key in tqdm(self.data.keys()):
            for period in self.seasonPeriods:
                if self.tag + key + '_v{}.png'.format(self.version) in os.listdir(self.folder + 'Seasonality/'):
                    continue
                seasonality = STL(self.data[key], period=period).fit()
                fig = plt.figure(figsize=[24, 16])
                plt.plot(range(len(self.data)), self.data[key])
                plt.plot(range(len(self.data)), seasonality)
                plt.title(key + ', period=' + str(period))
                fig.savefig(self.folder + 'Seasonality/' + self.tag + str(period)+'/'+key +
                            '_v{}.png'.format(self.version), format='png', dpi=300)
                plt.close()

    def co_linearity(self):
        if self.plotCoLinearity:
            # Create folder
            if not os.path.exists(self.folder + 'CoLinearity/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'CoLinearity/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'MinimumRepresentation.png' in \
                    os.listdir(self.folder + 'CoLinearity/v{}/'.format(self.version)):
                return

            # Plot threshold matrix
            threshold = 0.95
            fig = plt.figure(figsize=[24, 16])
            plt.title('Co-linearity matrix, threshold {:.2f}'.format(threshold))
            sns.heatmap(abs(self.data.corr()) < threshold, annot=False, cmap='Greys')
            fig.savefig(self.folder + 'CoLinearity/v{}/'.format(self.version) + self.tag + 'Matrix.png',
                        format='png', dpi=300)

            # Minimum representation
            corr_mat = self.data.corr().abs()
            upper = corr_mat.where(np.triu(np.ones(corr_mat.shape), k=1).astype('bool'))
            col_drop = [column for column in upper.columns if any(upper[column] > 0.95)]
            minimal_rep = self.data.drop(self.data[col_drop], axis=1)
            fig = plt.figure(figsize=[24, 16])
            sns.heatmap(abs(minimal_rep.corr()) < threshold, annot=False, cmap='Greys')
            fig.savefig(self.folder + 'CoLinearity/v{}/'.format(self.version) + self.tag + 'Minimum_Representation.png',
                        format='png', dpi=300)

    def differencing(self):
        if self.plotDifferencing:
            # Create folder
            if not os.path.exists(self.folder + 'Lags/'):
                os.mkdir(self.folder + 'Lags/')

            # Skip if existing
            if self.tag + 'Variance.png' in os.listdir(self.folder + 'Lags/'):
                return

            # Setup
            max_lags = 4
            var_vec = np.zeros((max_lags, len(self.data.keys())))
            diff_data = self.data / np.sqrt(self.data.var())

            # Calculate variance per lag
            for i in range(max_lags):
                var_vec[i, :] = diff_data.var()
                diff_data = diff_data.diff(1)[1:]

            # Plot
            fig = plt.figure(figsize=[24, 16])
            plt.title('Variance for different lags')
            plt.plot(var_vec)
            plt.xlabel('Lag')
            plt.yscale('log')
            plt.ylabel('Average variance')
            fig.savefig(self.folder + 'Lags/' + self.tag + 'Variance.png', format='png', dpi=300)

    def complete_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/ACF/'):
                os.makedirs(self.folder + 'Correlation/ACF/')

            # Difference data
            diff_data = copy.copy(self.data)
            for i in range(self.differ):
                diff_data = diff_data.diff(1)[1:]

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + 'Correlation/ACF/'):
                    continue

                # Plot
                fig = plot_acf(diff_data[key], fft=True)
                plt.title(key)
                fig.savefig(self.folder + 'Correlation/ACF/' + self.tag + key + '_differ_' + str(self.differ) +
                            '_v{}.png'.format(self.version), format='png', dpi=300)
                plt.close()

    def partial_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/PACF/'):
                os.makedirs(self.folder + 'Correlation/PACF/')

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + 'EDA/Correlation/PACF/'):
                    continue

                # Plot
                try:
                    fig = plot_pacf(self.data[key])
                    fig.savefig(self.folder + 'EDA/Correlation/PACF/' + self.tag + key + '_differ_' +
                                str(self.differ) + '_v{}.png'.format(self.version), format='png', dpi=300)
                    plt.title(key)
                    plt.close()
                except Exception as e:
                    # todo find exception
                    raise ValueError(e)

    def cross_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/Cross/'):
                os.makedirs(self.folder + 'Correlation/Cross/')

            # Prepare
            folder = 'Correlation/Cross/'
            y = self.Y.to_numpy().reshape((-1))

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + folder):
                    continue

                # Plot
                try:
                    fig = plt.figure(figsize=[24, 16])
                    plt.xcorr(self.data[key], y, maxlags=self.lags)
                    plt.title(key)
                    fig.savefig(self.folder + folder + self.tag + key + '_differ_' + str(self.differ) +
                                '_v{}.png'.format(self.version), format='png', dpi=300)
                    plt.close()
                except Exception as e:
                    raise ValueError(e)

    def scatters(self):
        if self.plotScatterPlots:
            # Create folder
            if not os.path.exists(self.folder + 'Scatters/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'Scatters/v{}/'.format(self.version))

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if '{}{}.png'.format(self.tag, key) in os.listdir(self.folder + 'Scatters/v{}/'.format(self.version)):
                    continue

                # Plot
                fig = plt.figure(figsize=[24, 16])
                plt.scatter(self.Y, self.data[key], alpha=0.2)
                plt.ylabel(key)
                plt.xlabel('Output')
                plt.title('Scatter Plot ' + key + ' - Output')
                fig.savefig(self.folder + 'Scatters/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def shap(self, args=None):
        args = args if args is not None else {}
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'Features/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'SHAP.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Calculate SHAP values
            import shap
            shap_values = shap.TreeExplainer(model).shap_values(self.data)

            # Plot
            fig = plt.figure(figsize=[8, 32])
            plt.subplots_adjust(left=0.4)
            shap.summary_plot(shap_values, self.data, plot_type='bar')
            fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'SHAP.png', format='png', dpi=300)

    def feature_ranking(self, **args):
        args = args if args is not None else {}
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
                os.mkdir(self.folder + 'Features/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'RF.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Plot
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ind = np.argsort(model.feature_importances_)
            plt.barh(list(self.data.keys()[ind])[-15:], width=model.feature_importances_[ind][-15:],
                     color='#2369ec')
            fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'RF.png', format='png', dpi=300)
            plt.close()

            # Store results
            results = pd.DataFrame({'x': self.data.keys(), 'score': model.feature_importances_})
            results.to_csv(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'RF.csv')

    def predictive_power_score(self):
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
                os.mkdir(self.folder + 'Features/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'PPScore.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
                return

            # Calculate PPS
            data = self.data.copy()
            if isinstance(self.Y, pd.core.series.Series):
                data.loc[:, 'target'] = self.Y
            elif isinstance(self.Y, pd.DataFrame):
                data.loc[:, 'target'] = self.Y.loc[:, self.Y.keys()[0]]
            pp_score = ppscore.predictors(data, 'target').sort_values('ppscore')

            # Plot
            fig, ax = plt.subplots(figsize=[4, 6])
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            plt.barh(pp_score['x'][-15:], width=pp_score['ppscore'][-15:], color='#2369ec')
            fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'Ppscore.png', format='png', dpi=400)
            plt.close()

            # Store results
            pp_score.to_csv(self.folder + 'Features/v{}/pp_score.csv'.format(self.version))



import copy
import time
import optuna
import warnings
import numpy as np
import pandas as pd
from datetime import datetime
from sklearn.model_selection import KFold
from sklearn.metrics import SCORERS


class OptunaGridSearch:
    # todo remove params from arg

    def __init__(self, model, params=None, cv=KFold(n_splits=3), scoring='accuracy', verbose=0, timeout=3600,
                 candidates=250):
        """
        Wrapper for Optuna Grid Search. Takes any model support by Amplo.AutoML.Modelling.
        The parameter search space is predefined for each model.

        Parameters
        ----------
        model obj: Model object to optimize
        params : deprecated
        cv obj: Scikit CV object
        scoring str: From Scikits Scorers*
        verbose int: How much to print
        timeout int: Time limit of optimization
        candidates int: Candidate limits to evaluate
        """
        self.model = model
        if hasattr(model, 'is_fitted'):
            assert not model.is_fitted(), 'Model already fitted'
        self.params = params
        self.cv = cv
        self.scoring = SCORERS[scoring] if isinstance(scoring, str) else scoring
        self.verbose = verbose
        self.timeout = timeout
        self.nTrials = candidates
        self.x, self.y = None, None
        self.binary = True
        self.samples = None

        # Model specific settings
        if type(self.model).__name__ == 'LinearRegression':
            self.nTrials = 1

        # Input tests
        assert model is not None, 'Need to provide a model'
        if scoring is None:
            if 'Classifier' in type(model).__name__:
                self.scoring = SCORERS['accuracy']
            elif 'Regressor' in type(model).__name__:
                self.scoring = SCORERS['neg_mean_squared_error']
            else:
                raise ValueError('Model mode unknown')

    def get_params(self, trial):
        # todo support suggest_loguniform
        if type(self.model).__name__ == 'LinearRegression':
            return {}
        elif type(self.model).__name__ == 'Lasso' or \
                'Ridge' in type(self.model).__name__:
            return {
                'alpha': trial.suggest_uniform('alpha', 0, 10),
            }
        elif 'SV' in type(self.model).__name__:
            return {
                'gamma': trial.suggest_categorical('gamma', ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1]),
                'C': trial.suggest_uniform('C', 0, 10),
            }
        elif 'KNeighbors' in type(self.model).__name__:
            return {
                'n_neighbors': trial.suggest_int('n_neighbors', 5, min(50, int(self.samples / 10))),
                'weights': trial.suggest_categorical('weights', ['uniform', 'distance']),
                'leaf_size': trial.suggest_int('leaf_size', 1, min(100, int(self.samples / 10))),
            }

        # Regressors
        elif type(self.model).__name__ == 'DecisionTreeRegressor':
            return {
                'criterion': trial.suggest_categorical('criterion', ['mse', 'friedman_mse', 'mae', 'poisson']),
                'max_depth': trial.sugest_int('max_depth', 3, min(25, int(np.log2(self.samples)))),
            }
        elif type(self.model).__name__ == 'BaggingRegressor':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 10, 250),
                'max_samples': trial.suggest_uniform('max_samples', 0.5, 1),
                'max_features': trial.suggest_uniform('max_features', 0.5, 1),
            }
        elif type(self.model).__name__ == 'CatBoostRegressor':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 500, 2000),
                'loss_function': trial.suggest_categorical('loss', ['MAE', 'RMSE']),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'l2_leaf_reg': trial.suggest_uniform('l2_leaf_reg', 0, 10),
                'depth': trial.suggest_int('depth', 3, min(10, int(np.log2(self.samples)))),
                'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 1, min(1000, int(self.samples / 10))),
                'max_leaves': trial.suggest_int('max_leaves', 10, 250),
            }
        elif type(self.model).__name__ == 'GradientBoostingRegressor':
            return {
                'loss': trial.suggest_categorical('loss', ['ls', 'lad', 'huber']),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'max_depth': trial.suggest_int('max_depth', 3, min(10, int(np.log2(self.samples)))),
                'n_estimators': trial.suggest_int('n_estimators', 100, 1000),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'max_features': trial.suggest_uniform('max_features', 0.5, 1),
                'subsample': trial.suggest_uniform('subsample', 0.5, 1),
            }
        elif type(self.model).__name__ == 'HistGradientBoostingRegressor':
            return {
                'loss': trial.suggest_categorical('loss', ['least_squares', 'least_absolute_deviation']),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'max_iter': trial.suggest_int('max_iter', 100, 250),
                'max_leaf_nodes': trial.suggest_int('max_leaf_nodes', 30, 150),
                'max_depth': trial.suggest_int('max_depth', 3, min(10, int(np.log2(self.samples)))),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'l2_regularization': trial.suggest_uniform('l2_regularization', 0, 10),
                'max_bins': trial.suggest_int('max_bins', 100, 255),
                'early_stopping': [True],
            }
        elif type(self.model).__name__ == 'RandomForestRegressor':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 50, 1000),
                'criterion': trial.suggest_categorical('criterion', ['mse', 'mae']),
                'max_depth': trial.suggest_int('max_depth', 3, min(15, int(np.log2(self.samples)))),
                'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt']),
                'min_samples_split': trial.suggest_int('min_samples_split', 2, 50),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'bootstrap': trial.suggest_categorical('bootstrap', [True, False]),
            }
        elif type(self.model).__name__ == 'XGBRegressor':
            param = {
                "objective": trial.suggest_categorical('objective', ['reg:squarederror', 'reg:squaredlogerror',
                                                                     'reg:logistic']),
                "eval_metric": "rmse",
                "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear", "dart"]),
                "lambda": trial.suggest_loguniform("lambda", 1e-8, 1.0),
                "alpha": trial.suggest_loguniform("alpha", 1e-8, 1.0),
                "callbacks": optuna.integration.XGBoostPruningCallback(trial, "validation-logloss"),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
            }
            if param["booster"] == "gbtree" or param["booster"] == "dart":
                param["max_depth"] = trial.suggest_int("max_depth", 1, min(10, int(np.log2(self.samples))))
                param["eta"] = trial.suggest_loguniform("eta", 1e-8, 1.0)
                param["gamma"] = trial.suggest_loguniform("gamma", 1e-8, 1.0)
                param["grow_policy"] = trial.suggest_categorical("grow_policy", ["depthwise", "lossguide"])
            if param["booster"] == "dart":
                param["sample_type"] = trial.suggest_categorical("sample_type", ["uniform", "weighted"])
                param["normalize_type"] = trial.suggest_categorical("normalize_type", ["tree", "forest"])
                param["rate_drop"] = trial.suggest_loguniform("rate_drop", 1e-8, 1.0)
                param["skip_drop"] = trial.suggest_loguniform("skip_drop", 1e-8, 1.0)
            return param
        elif type(self.model).__name__ == 'LGBMRegressor':
            return {
                'num_leaves': trial.suggest_int('num_leaves', 10, 150),
                'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 1, min(1000, int(self.samples / 10))),
                'min_sum_hessian_in_leaf': trial.suggest_uniform('min_sum_hessian_in_leaf', 1e-3, 0.5),
                'subsample': trial.suggest_uniform('subsample', 0.5, 1),
                'colsample_bytree': trial.suggest_uniform('colsample_bytree', 0, 1),
                'reg_alpha': trial.suggest_uniform('reg_alpha', 0, 1),
                'reg_lambda': trial.suggest_uniform('reg_lambda', 0, 1),
            }

        # Classifiers
        elif type(self.model).__name__ == 'BaggingClassifier':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 10, 250),
                'max_samples': trial.suggest_uniform('max_samples', 0.5, 1),
                'max_features': trial.suggest_uniform('max_features', 0.5, 1),
            }
        elif type(self.model).__name__ == 'CatBoostClassifier':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 500, 2000),
                "verbose": 0,
                'early_stopping_rounds': 100,
                'od_pval': 1e-5,
                'loss_function': 'Logloss' if self.binary else 'MultiClass',
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'l2_leaf_reg': trial.suggest_uniform('l2_leaf_reg', 0, 10),
                'depth': trial.suggest_int('depth', 1, min(10, int(np.log2(self.samples)))),
                'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 1, min(1000, int(self.samples / 10))),
                'grow_policy': trial.suggest_categorical('grow_policy', ['SymmetricTree', 'Depthwise', 'Lossguide']),
            }
        elif type(self.model).__name__ == 'GradientBoostingClassifier':
            return {
                'loss': trial.suggest_categorical('loss', ['deviance', 'exponential']),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'max_depth': trial.suggest_int('max_depth', 3, min(15, int(np.log2(self.samples)))),
                'n_estimators': trial.suggest_int('n_estimators', 100, 250),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'max_features': trial.suggest_uniform('max_features', 0.5, 1),
                'subsample': trial.suggest_uniform('subsample', 0.5, 1),
            }
        elif type(self.model).__name__ == 'HistGradientBoostingClassifier':
            return {
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
                'max_iter': trial.suggest_int('max_iter', 100, 1000),
                'max_leaf_nodes': trial.suggest_int('max_leaf_nodes', 30, 150),
                'max_depth': trial.suggest_int('max_depth', 3, min(10, int(np.log2(self.samples)))),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'l2_regularization': trial.suggest_uniform('l2_regularization', 0, 10),
                'max_bins': trial.suggest_int('max_bins', 100, 255),
                'early_stopping': [True],
            }
        elif type(self.model).__name__ == 'RandomForestClassifier':
            return {
                'n_estimators': trial.suggest_int('n_estimators', 50, 1000),
                'criterion': trial.suggest_categorical('criterion', ['gini', 'entropy']),
                'max_depth': trial.suggest_int('max_depth', 3, min(15, int(np.log2(self.samples)))),
                'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt']),
                'min_samples_split': trial.suggest_int('min_samples_split', 2, 50),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, min(1000, int(self.samples / 10))),
                'bootstrap': trial.suggest_categorical('bootstrap', [True, False]),
            }
        elif type(self.model).__name__ == 'XGBClassifier':
            param = {
                "objective": "binary:logistic" if self.binary else 'multi:softprob',
                "eval_metric": "logloss",
                "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear", "dart"]),
                "lambda": trial.suggest_loguniform("lambda", 1e-8, 1.0),
                "alpha": trial.suggest_loguniform("alpha", 1e-8, 1.0),
                "callbacks": optuna.integration.XGBoostPruningCallback(trial, "validation-logloss"),
                'learning_rate': trial.suggest_loguniform('learning_rate', 0.001, 0.5),
            }
            if param["booster"] == "gbtree" or param["booster"] == "dart":
                param["max_depth"] = trial.suggest_int("max_depth", 1, min(10, int(np.log2(self.samples))))
                param["eta"] = trial.suggest_loguniform("eta", 1e-8, 1.0)
                param["gamma"] = trial.suggest_loguniform("gamma", 1e-8, 1.0)
                param["grow_policy"] = trial.suggest_categorical("grow_policy", ["depthwise", "lossguide"])
            if param["booster"] == "dart":
                param["sample_type"] = trial.suggest_categorical("sample_type", ["uniform", "weighted"])
                param["normalize_type"] = trial.suggest_categorical("normalize_type", ["tree", "forest"])
                param["rate_drop"] = trial.suggest_loguniform("rate_drop", 1e-8, 1.0)
                param["skip_drop"] = trial.suggest_loguniform("skip_drop", 1e-8, 1.0)
            return param
        elif type(self.model).__name__ == 'LGBMClassifier':
            return {
                "objective": "binary" if self.binary else 'multiclass',
                "metric": trial.suggest_categorical("metric", ['binary_error', 'auc', 'average_precision',
                                                               'binary_logloss']) if self.binary else
                trial.suggest_categorical('metric', ['multi_error', 'multi_logloss', 'auc_mu']),
                "verbosity": -1,
                "boosting_type": "gbdt",
                "lambda_l1": trial.suggest_loguniform("lambda_l1", 1e-8, 10.0),
                "lambda_l2": trial.suggest_loguniform("lambda_l2", 1e-8, 10.0),
                "num_leaves": trial.suggest_int("num_leaves", 2, 256),
                "feature_fraction": trial.suggest_uniform("feature_fraction", 0.4, 1.0),
                "bagging_fraction": trial.suggest_uniform("bagging_fraction", 0.4, 1.0),
                "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
                "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 1, min(1000, int(self.samples / 10))),
                'callbacks': optuna.integration.LightGBMPruningCallback(trial, "neg_log_loss", "valid_1"),
            }
        else:
            # Raise error if nothing is returned
            warnings.warn('Hyper parameter tuning not implemented for {}'.format(type(self.model).__name__))
            return {}

    def fit(self, x, y):
        if isinstance(y, pd.DataFrame):
            assert len(y.keys()) == 1, 'Multiple target columns not supported.'
            y = y[y.keys()[0]]
        assert isinstance(x, pd.DataFrame), 'X should be Pandas DataFrame'
        assert isinstance(y, pd.Series), 'Y should be Pandas Series or DataFrame'

        # Set mode
        self.binary = y.nunique() == 2
        self.samples = len(y)

        # Store
        self.x, self.y = x, y

        # Set up study
        study = optuna.create_study(sampler=optuna.samplers.TPESampler(), direction='maximize')
        study.optimize(self.objective, timeout=self.timeout, n_trials=self.nTrials)

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame({
            'date': datetime.today().strftime('%d %b %y'),
            'model': type(self.model).__name__,
            'params': [x.params for x in study.get_trials()],
            'mean_objective': optuna_results['value'],
            'std_objective': optuna_results['user_attrs_std_value'],
            'worst_case': optuna_results['value'] - optuna_results['user_attrs_std_value'],
            'mean_time': optuna_results['user_attrs_mean_time'],
            'std_time': optuna_results['user_attrs_std_time']
        })
        return results

    def objective(self, trial):
        # Metrics
        scores = []
        times = []
        master = copy.deepcopy(self.model)

        # Cross Validation
        for t, v in self.cv.split(self.x, self.y):
            # Split data
            xt, xv, yt, yv = self.x.iloc[t], self.x.iloc[v], self.y.iloc[t], self.y.iloc[v]

            # Train model
            t_start = time.time()
            model = copy.deepcopy(master)
            model.set_params(**self.get_params(trial))
            model.fit(xt, yt)

            # Results
            scores.append(self.scoring(model, xv, yv))
            times.append(time.time() - t_start)

        trial.set_user_attr('mean_time', np.mean(times))
        trial.set_user_attr('std_time', np.std(times))
        trial.set_user_attr('std_value', np.std(scores))
        return np.mean(scores)



import unittest
import numpy as np
import pandas as pd
from sklearn.metrics import SCORERS
from sklearn.base import clone
from sklearn.datasets import make_classification
from Amplo.Classifiers import XGBClassifier


class TestXGBClassifier(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        x, y = make_classification(n_classes=5, n_informative=15)
        cls.x, cls.y = pd.DataFrame(x), pd.Series(y)

    def test_set_params(self):
        model = XGBClassifier()
        model.set_params(**{'depth': 200})

    def test_get_params(self):
        model = XGBClassifier()
        model.get_params()

    def test_pd_fit(self):
        model = XGBClassifier()
        model.fit(self.x, self.y)
        prediction = model.predict(self.x)
        assert not np.isnan(prediction).any(), 'NaN in prediction: {}'.format(prediction)

    def test_np_fit(self):
        model = XGBClassifier()
        model.fit(self.x.to_numpy(), self.y.to_numpy())
        prediction = model.predict(self.x.to_numpy())
        assert not np.isnan(prediction).any(), 'NaN in prediction: {}'.format(prediction)

    def test_trained_attr(self):
        model = XGBClassifier()
        assert hasattr(model, 'trained'), 'Attribute missing'
        assert model.trained is False
        model.fit(self.x, self.y)
        assert model.trained is True

    def test_predict(self):
        model = XGBClassifier()
        model.fit(self.x, self.y)
        prediction = model.predict(self.x)

        # Tests
        assert not np.isnan(prediction).any(), 'NaN in prediction: {}'.format(prediction)
        assert len(prediction.shape) == 1
        assert np.allclose(prediction.astype('int'), prediction)

    def test_probability(self):
        model = XGBClassifier()
        model.fit(self.x, self.y)
        prediction = model.predict_proba(self.x)

        assert model.binary is False
        assert not np.isnan(prediction).any(), 'NaN in prediction: {}'.format(prediction)
        assert len(prediction.shape) == 2
        assert prediction.shape[1] == len(np.unique(self.y))
        assert np.allclose(np.sum(prediction, axis=1), 1)

    def test_cloneable(self):
        model = XGBClassifier()
        params = {'max_depth': 10}
        cloned = clone(clone(model).set_params(**params))
        cloned.fit(self.x, self.y)
        prediction = cloned.predict_proba(self.x)

        assert len(prediction.shape) != 1
        assert np.allclose(np.sum(prediction, axis=1), 1)

    def test_scorer(self):
        model = XGBClassifier()
        model.fit(self.x, self.y)
        SCORERS['neg_log_loss'](model, self.x, self.y)
        SCORERS['accuracy'](model, self.x, self.y)
        SCORERS['f1_micro'](model, self.x, self.y)

    def test_binary(self):
        model = XGBClassifier()
        x, y = make_classification()
        model.fit(x, y)
        prediction = model.predict(x)
        proba = model.predict_proba(x)
        SCORERS['neg_log_loss'](model, x, y)
        SCORERS['accuracy'](model, x, y)
        SCORERS['f1_micro'](model, x, y)



import os
import copy
import time
import joblib
import numpy as np
import pandas as pd
from datetime import datetime
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from sklearn.experimental import enable_hist_gradient_boosting
from ..Classifiers.CatBoostClassifier import CatBoostClassifier
from ..Classifiers.XGBClassifier import XGBClassifier
from ..Classifiers.LGBMClassifier import LGBMClassifier
from sklearn import linear_model
from sklearn import ensemble
from sklearn import svm
from sklearn import metrics


class Modelling:

    def __init__(self,
                 mode='regression',
                 shuffle=False,
                 n_splits=3,
                 objective='accuracy',
                 samples=None,
                 folder='',
                 dataset='set_0',
                 store_models=False,
                 store_results=True):
        """
        Runs various regression or classification models.
        Includes:
        - Scikit's Linear Model
        - Scikit's Random Forest
        - Scikit's Bagging
        - Scikit's GradientBoosting
        - Scikit's HistGradientBoosting
        - DMLC's XGBoost
        - Catboost's Catboost
        - Microsoft's LightGBM

        Parameters
        ----------
        mode str: 'regression' or 'classification'
        shuffle bool: Whether to shuffle samples for training / validation
        n_splits int: Number of cross-validation splits
        objective str: Performance metric from SciKit Scorers*
        samples int: Samples in dataset, does not need to be specified but useful for calling return_models()
        folder str: Folder to store models and / or results
        dataset str: Name of feature set, documentation purpose
        store_models bool: Whether to store the trained models
        store_results bool:Whether to store the results

        * https://scikit-learn.org/stable/modules/model_evaluation.html
        """
        # Test
        assert mode in ['classification', 'regression'], 'Unsupported mode'
        assert isinstance(shuffle, bool)
        assert isinstance(n_splits, int)
        assert 2 < n_splits < 10, 'Reconsider your number of splits'
        assert isinstance(objective, str)
        assert objective in metrics.SCORERS.keys(), \
            'Pick scorer from sklearn.metrics.SCORERS: \n{}'.format(list(metrics.SCORERS.keys()))
        assert isinstance(samples, int) or samples is None
        assert isinstance(folder, str)
        assert isinstance(dataset, str)
        assert isinstance(store_models, bool)
        assert isinstance(store_results, bool)

        # Parameters
        self.objective = objective
        self.scoring = metrics.SCORERS[objective]
        self.mode = mode
        self.shuffle = shuffle
        self.cvSplits = n_splits
        self.samples = samples
        self.dataset = str(dataset)
        self.storeResults = store_results
        self.storeModels = store_models
        self.results = pd.DataFrame(columns=['date', 'model', 'dataset', 'params', 'mean_objective', 'std_objective',
                                             'mean_time', 'std_time'])

        # Folder
        self.folder = folder if len(folder) == 0 or folder[-1] == '/' else folder + '/'
        if (store_results or store_models) and self.folder != '':
            if not os.path.exists(self.folder):
                os.makedirs(self.folder)

        self.needsProba = False  # Whether scorer requires needs_proba attr
        if 'True' in self.scoring._factory_args():
            self.needsProba = True

    def fit(self, x, y):
        # Copy number of samples
        self.samples = len(y)

        # Regression
        if self.mode == 'regression':
            cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            return self._fit(x, y, cv)

        # Classification
        if self.mode == 'classification':
            cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            return self._fit(x, y, cv)

    def return_models(self):
        models = []

        # All classifiers
        if self.mode == 'classification':
            # The thorough ones
            if self.samples < 25000:
                if not self.needsProba:
                    models.append(linear_model.RidgeClassifier())
                    models.append(svm.SVC(kernel='rbf'))
                models.append(ensemble.BaggingClassifier())
                models.append(ensemble.GradientBoostingClassifier())
                models.append(XGBClassifier())

            # The efficient ones
            else:
                if not self.needsProba:
                    models.append(linear_model.RidgeClassifier())
                models.append(ensemble.HistGradientBoostingClassifier())
                models.append(LGBMClassifier())

            # And the multifaceted ones
            models.append(CatBoostClassifier())
            models.append(ensemble.RandomForestClassifier())

        elif self.mode == 'regression':
            # The thorough ones
            if self.samples < 25000:
                if not self.needsProba:
                    models.append(linear_model.LinearRegression())
                    models.append(svm.SVR(kernel='rbf'))
                models.append(ensemble.BaggingRegressor())
                models.append(ensemble.GradientBoostingRegressor())
                # todo XG Boost

            # The efficient ones
            else:
                if not self.needsProba:
                    models.append(linear_model.LinearRegression())
                models.append(ensemble.HistGradientBoostingRegressor())
                # todo LGBM

            # And the multifaceted ones
            # todo Catboost
            models.append(ensemble.RandomForestRegressor())

        return models

    def _fit(self, x, y, cross_val):
        # Convert to NumPy
        x = np.array(x)
        y = np.array(y).ravel()

        # Data
        print('[Modelling] Splitting data (shuffle=%s, splits=%i, features=%i)' %
              (str(self.shuffle), self.cvSplits, len(x[0])))

        if self.storeResults and 'Initial_Models.csv' in os.listdir(self.folder):
            self.results = pd.read_csv(self.folder + 'Initial_Models.csv')
            for i in range(len(self.results)):
                self.print_results(self.results.iloc[i])

        else:

            # Models
            self.models = self.return_models()

            # Loop through models
            for master_model in self.models:

                # Time & loops through Cross-Validation
                val_score = []
                train_score = []
                train_time = []
                for t, v in cross_val.split(x, y):
                    t_start = time.time()
                    xt, xv, yt, yv = x[t], x[v], y[t], y[v]
                    model = copy.copy(master_model)
                    model.fit(xt, yt)
                    val_score.append(self.scoring(model, xv, yv))
                    train_score.append(self.scoring(model, xt, yt))
                    train_time.append(time.time() - t_start)

                # Append results
                result = {
                    'date': datetime.today().strftime('%d %b %y'),
                    'model': type(model).__name__,
                    'dataset': self.dataset,
                    'params': model.get_params(),
                    'mean_objective': np.mean(val_score),
                    'std_objective': np.std(val_score),
                    'worst_case': np.mean(val_score) - np.std(val_score),
                    'mean_time': np.mean(train_time),
                    'std_time': np.std(train_time)
                }
                self.results = self.results.append(result, ignore_index=True)
                self.print_results(result)

                # Store model
                if self.storeModels:
                    joblib.dump(model, self.folder + type(model).__name__ + '_{:.4f}.joblib'.format(np.mean(val_score)))

            # Store CSV
            if self.storeResults:
                self.results.to_csv(self.folder + 'Initial_Models.csv')

        # Return results
        return self.results

    def print_results(self, result):
        print('[Modelling] {} {}: {:.4f} \u00B1 {:.4f}, training time: {:.1f} s'.format(
            result['model'].ljust(30), self.objective, result['mean_objective'],
            result['std_objective'], result['mean_time']))



import pandas as pd


class DataSampler:

    def __init__(self):
        '''
        Trial and error various strategies:
        - Undersample all to minority
        - Upsample all to majority
        - Combination of the two, with say 1k, 5k, 25k, 50k, 75k and 100k samples
        This sensitivity is problem dependent, but has major consequence for hyper
        parameter optimization. Need to get it right.
        https://imbalanced-learn.org/stable/references/index.html#api
        '''
        pass



import unittest
import warnings
warnings.filterwarnings('ignore')
import pandas as pd
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression
from Amplo.GridSearch import HalvingGridSearch
from Amplo.AutoML import Modelling


class TestHalvingGridSearch(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        x, y = make_classification()
        cls.cx = pd.DataFrame(x)
        cls.cy = pd.Series(y)
        x, y = make_regression()
        cls.rx = pd.DataFrame(x)
        cls.ry = pd.Series(y)

    def test_small_regression(self):
        models = Modelling(mode='regression', objective='r2', samples=100).return_models()
        for model in models:
            grid_search = HalvingGridSearch(model,
                                            cv=KFold(n_splits=3),
                                            verbose=0,
                                            candidates=2,
                                            scoring='r2')
            results = grid_search.fit(self.rx, self.ry)

            # Tests
            assert isinstance(results, pd.DataFrame), 'Not a DF {}'.format(type(model).__name__)
            assert all([x in results.keys() for x in ['worst_case', 'mean_objective', 'std_objective', 'params',
                                                      'mean_time', 'std_time']]), 'Keys missing {}'.format(type(model)
                                                                                                           .__name__)
            assert len(results) > 0, 'Empty results {}'.format(type(model).__name__)

    def test_big_regression(self):
        models = Modelling(mode='regression', objective='r2', samples=50000).return_models()
        for model in models:
            grid_search = HalvingGridSearch(model,
                                            cv=KFold(n_splits=3),
                                            verbose=0,
                                            candidates=2,
                                            scoring='r2')
            results = grid_search.fit(self.rx, self.ry)

            # Tests
            assert isinstance(results, pd.DataFrame), 'Not a DF {}'.format(type(model).__name__)
            assert all([x in results.keys() for x in ['worst_case', 'mean_objective', 'std_objective', 'params',
                                                      'mean_time', 'std_time']]), 'Keys missing {}'.format(type(model)
                                                                                                           .__name__)
            assert len(results) > 0, 'Empty results {}'.format(type(model).__name__)

    def test_small_classification(self):
        models = Modelling(mode='classification', objective='accuracy', samples=100).return_models()
        for model in models:
            assert hasattr(model, 'fit'), 'Has no fit attribute'
            grid_search = HalvingGridSearch(model,
                                            cv=StratifiedKFold(n_splits=3),
                                            verbose=0,
                                            candidates=2,
                                            scoring='accuracy')
            results = grid_search.fit(self.cx, self.cy)

            # Tests
            assert isinstance(results, pd.DataFrame), 'Not a DF {}'.format(type(model).__name__)
            assert all([x in results.keys() for x in ['worst_case', 'mean_objective', 'std_objective', 'params',
                                                      'mean_time', 'std_time']]), 'Keys missing {}'.format(type(model)
                                                                                                           .__name__)
            assert len(results) > 0, 'Empty results {}'.format(type(model).__name__)

    def test_big_classification(self):
        models = Modelling(mode='classification', objective='accuracy', samples=50000).return_models()
        for model in models:
            grid_search = HalvingGridSearch(model,
                                            cv=StratifiedKFold(n_splits=3),
                                            verbose=0,
                                            candidates=2,
                                            scoring='accuracy')
            results = grid_search.fit(self.cx, self.cy)

            # Tests
            assert isinstance(results, pd.DataFrame), 'Not a DF {}'.format(type(model).__name__)
            assert all([x in results.keys() for x in ['worst_case', 'mean_objective', 'std_objective', 'params',
                                                      'mean_time', 'std_time']]), 'Keys missing {}'.format(type(model)
                                                                                                           .__name__)
            assert len(results) > 0, 'Empty results {}'.format(type(model).__name__)



import os
import pytest
import shutil
import unittest
from sklearn.datasets import load_iris
from sklearn.datasets import load_boston
from Amplo.AutoML import DataExploring


class TestDataExploring(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.class_x, cls.class_y = load_iris(return_X_y=True, as_frame=True)
        cls.reg_x, cls.reg_y = load_boston(return_X_y=True)

    def test_regression(self):
        eda = DataExploring(self.reg_x, y=self.reg_y, mode='regression', folder='tmp/')
        eda.run()

    def test_classification(self):
        eda = DataExploring(self.class_x, y=self.class_y, mode='classification', folder='tmp/')
        eda.run()


@pytest.fixture(scope="session", autouse=True)
def teardown():
    yield
    if os.path.exists('tmp'):
        shutil.rmtree('tmp')



import os
import copy
import numpy as np
from sklearn import metrics
import matplotlib.pyplot as plt
from sklearn.model_selection import StratifiedKFold
from .BinaryDocumenting import BinaryDocumenting


class MultiDocumenting(BinaryDocumenting):

    def __init__(self, pipeline):
        super().__init__(pipeline)

    def analyse(self):
        # Initiating
        f1_score = np.zeros((self.p.cvSplits, self.p.n_classes))
        log_loss = np.zeros(self.p.cvSplits)
        avg_acc = np.zeros(self.p.cvSplits)
        cm = np.zeros((self.p.cvSplits, self.p.n_classes, self.p.n_classes))

        # Modelling
        self.cv = StratifiedKFold(n_splits=self.p.cvSplits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            xt, xv, yt, yv = self.x[t], self.x[v], self.y[t].reshape((-1)), self.y[v].reshape((-1))
            model = copy.copy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))
            cm[i] = metrics.confusion_matrix(predictions, yv)

            # Metrics
            f1_score[i] = metrics.f1_score(yv, predictions, average=None)
            avg_acc[i] = metrics.accuracy_score(yv, predictions)
            if hasattr(model, 'predict_proba'):
                probabilities = model.predict_proba(xv)
                log_loss[i] = metrics.log_loss(yv, probabilities)

        # Result statistics
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0)
        stds = np.std(cm / totals * 100, axis=0)

        # Store
        self.metrics = {
            'F1 Score': [np.mean(f1_score), np.std(f1_score)],
            'Accuracy': [np.mean(avg_acc), np.std(avg_acc)],
        }
        self.confusion_matrix = {
            'means': means,
            'stds': stds,
        }

        # Print
        print('F1 scores:')
        print(''.join([' Class {} |'.format(i) for i in range(self.p.n_classes)]))
        print(''.join([' {:.2f} % '.ljust(11).format(f1) + '|' for f1 in np.mean(f1_score, axis=0)]))
        print('Average Accuracy: {:.2f} \u00B1 {:.2f} %'.format(np.mean(avg_acc), np.std(avg_acc)))
        if hasattr(model, 'predict_proba'):
            print('Log Loss:         {:.2f} \u00B1 {:.2f}'.format(np.mean(log_loss), np.std(log_loss)))
            self.metrics['Log Loss'] = [np.mean(log_loss), np.std(log_loss)]

        if not os.path.exists(self.p.mainDir + 'EDA/Features/v{}/RF.png'.format(self.p.version)):
            if not os.path.exists(self.p.mainDir + 'EDA/Features/v{}'.format(self.p.version)):
                os.makedirs(self.p.mainDir + 'EDA/Features/v{}/'.format(self.p.version))
            from sklearn.ensemble import RandomForestClassifier
            model = RandomForestClassifier()
            model.fit(self.p.x, self.p.y)
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ind = np.argsort(model.feature_importances_)
            plt.barh(list(self.p.x.keys()[ind])[-15:], width=model.feature_importances_[ind][-15:],
                     color='#2369ec')
            fig.savefig(self.p.mainDir + 'EDA/Features/v{}/RF.png'.format(self.p.version), format='png', dpi=200)

    def model_performance(self):
        self.ln(self.lh)
        self.add_h2('Model Performance')
        self.add_text('Model performance is analysed by various metrics. This model has been selected based on the {} '
                      'score.'.format(self.p.objective))

        # Metrics
        self.set_font('Helvetica', 'B', 12)
        self.ln(self.lh)
        self.cell(w=50, h=self.lh, txt='Metric', border='B', align='C')
        self.cell(w=50, h=self.lh, txt='Score', border='LB', align='C')
        self.set_font('Helvetica', '', 12)
        for k, v in self.metrics.items():
            self.ln(self.lh)
            self.cell(w=50, h=self.lh, txt=k, border='R', align='L')
            self.cell(w=50, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(v[0], v[1]), border='L', align='C')
        self.ln(self.lh * 3)

        # Confusion Matrix
        nc = self.p.n_classes
        cell_width = int((self.WIDTH - self.pm * 2) / (nc + 2))
        self.add_h3('Confusion Matrix')

        # First row
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=cell_width * 2, h=self.lh, txt='', align='L', border='R')
        self.cell(w=cell_width * nc, h=self.lh, txt='True Class', align='C')
        self.ln(self.lh)

        # Second Row
        self.cell(w=cell_width * 2, h=self.lh, txt='', align='L', border='B')
        for i in range(nc):
            self.cell(w=cell_width, h=self.lh, txt='Class {}'.format(i), align='C', border='BL')
        self.ln(self.lh)

        # Values
        for i in range(nc):
            # Set bold
            self.set_font('Helvetica', 'B', 12)
            # Prediction (only for first time)
            if i == 0:
                self.cell(w=cell_width, h=self.lh * nc, txt='Prediction', align='L')
            else:
                self.cell(w=cell_width, h=self.lh, txt='')
            # Class
            self.cell(w=cell_width, h=self.lh, txt='Class {}'.format(i), align='L', border='R')
            # Set normal
            self.set_font('Helvetica', '', 12)
            for j in range(nc):
                self.cell(w=cell_width, h=self.lh, txt='{:.1f}\u00B1{:.1f} %'.format(
                    self.confusion_matrix['means'][i][j], self.confusion_matrix['stds'][i][j]), align='C')
            self.ln(self.lh)

    def validation(self):
        self.ln(self.lh)
        self.check_new_page()
        self.add_h3('Validation Strategy')
        self.add_text("All experiments are cross-validated. This means that every time a model's "
                      "performance is evaluated, it's trained on one part of the data, and test on another. Therefore, "
                      "the model is always test against data it has not yet been trained for. This gives the best "
                      "approximation for real world (out of sample) performance. The current validation strategy used "
                      "is {}, with {} splits and {} shuffling the data.".format(
            type(self.cv).__name__, self.p.cvSplits, 'with' if self.p.shuffle else 'without'))



import numpy as np


class BaseClassifier:

    def __init__(self, default_params=None, **params):
        assert isinstance(default_params, dict), 'Provided default parameters not of type dict'
        assert isinstance(params, dict), 'Provided parameters not of type dict'
        self.defaultParams = default_params
        self.model = None
        self.hasPredictProba = False
        self.trained = False
        self.classes_ = None
        self.callbacks = None
        self.params = params if params is not None else self.defaultParams
        for key in [k for k in self.defaultParams if k not in self.params]:
            self.params[key] = self.defaultParams[key]

    def get_params(self):
        return self.model.get_params()

    def set_params(self, **params):
        self.model.set_params(**params)
        return self

    def predict(self, x):
        return self.model.predict(x)

    def predict_proba(self, x):
        if self.hasPredictProba:
            return self.model.predict_proba(x)
        else:
            raise AttributeError('{} has no predict_proba'.format(type(self.model).__name__))

    def score(self, x, y):
        return self.model.score(x, y)

    def fit(self, x, y):
        self.classes_ = np.unique(y)
        self.model.fit(x, y)
        self.trained = True
        return self



import unittest
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression
from Amplo.AutoML import FeatureProcessing
from Amplo.Utils import check_dataframe_quality


class TestDataProcessing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        x, y = make_classification()
        cls.classification = pd.DataFrame({'target': y})
        for i in range(x.shape[1]):
            cls.classification['Feature_{}'.format(i)] = x[:, i]
        x, y = make_regression()
        cls.regression = pd.DataFrame({'target': y})
        for i in range(x.shape[1]):
            cls.regression['Feature_{}'.format(i)] = x[:, i]

    def test_regression(self):
        fp = FeatureProcessing(max_lags=2, mode='regression', folder='tmp/')
        extract = fp.extract(self.regression.drop('target', axis=1), self.regression['target'])
        assert check_dataframe_quality(extract)

    def test_classification(self):
        fp = FeatureProcessing(max_lags=2, mode='classification', folder='tmp/')
        extract = fp.extract(self.classification.drop('target', axis=1), self.classification['target'])
        assert check_dataframe_quality(extract)



import os
import shutil
import pytest
import unittest
import pandas as pd
from sklearn.datasets import make_regression
from sklearn.datasets import make_classification
from Amplo import Pipeline


# class TestDocumenting(unittest.TestCase):

    # def test_binary(self):
    #     x, y = make_classification(n_classes=2, n_informative=15)
    #     x = pd.DataFrame(x, columns=['Feature_{}'.format(i) for i in range(x.shape[1])])
    #     x['target'] = y
    #     pipeline = Pipeline('target',
    #                         project='Tritium',
    #                         device='PK350',
    #                         issue='DC Domes',
    #                         mode='classification',
    #                         objective='neg_log_loss',
    #                         feature_timeout=5,
    #                         grid_search_iterations=1,
    #                         grid_search_time_budget=10,
    #                         grid_search_candidates=2,
    #                         plot_eda=False,
    #                         process_data=False,
    #                         document_results=True)
    #     pipeline.fit(x)
    #     assert len(os.listdir('AutoML/Documentation/v0')) != 0

    # def test_multi(self):
    #     x, y = make_classification(n_classes=5, n_informative=15)
    #     x = pd.DataFrame(x, columns=['Feature_{}'.format(i) for i in range(x.shape[1])])
    #     x['target'] = y
    #     pipeline = Pipeline('target',
    #                         project='Tritium',
    #                         device='PK350',
    #                         issue='DC Domes',
    #                         mode='classification',
    #                         objective='neg_log_loss',
    #                         feature_timeout=5,
    #                         grid_search_iterations=1,
    #                         grid_search_time_budget=10,
    #                         grid_search_candidates=2,
    #                         plot_eda=False,
    #                         process_data=False,
    #                         document_results=True)
    #     pipeline.fit(x)
    #     assert len(os.listdir('AutoML/Documentation/')) != 0
#
    # def test_regression(self):
    #     x, y = make_regression()
    #     x = pd.DataFrame(x, columns=['Feature_{}'.format(i) for i in range(x.shape[1])])
    #     x['target'] = y
    #     pipeline = Pipeline('target',
    #                         project='Tritium',
    #                         device='PK350',
    #                         issue='DC Domes',
    #                         mode='regression',
    #                         objective='r2',
    #                         feature_timeout=5,
    #                         grid_search_iterations=1,
    #                         grid_search_time_budget=10,
    #                         grid_search_candidates=2,
    #                         plot_eda=False,
    #                         process_data=False,
    #                         document_results=True)
    #     pipeline.fit(x)
    #     assert len(os.listdir('AutoML/Documentation/')) != 0
#
#
# @pytest.fixture(scope='session', autouse=True)
# def teardown():
#     yield
#     for i in range(3):
#         if os.path.exists('doc{}'.format(i)):
#             shutil.rmtree('doc{}'.format(i))



import re
import os
import inspect
import textwrap
import numpy as np
import pandas as pd
from ..Utils import clean_keys


class DataProcessing:
    # todo implement data type detector

    def __init__(self,
                 target=None,
                 num_cols=None,
                 date_cols=None,
                 cat_cols=None,
                 missing_values='interpolate',
                 outlier_removal='clip',
                 z_score_threshold=4,
                 folder='',
                 version=1,
                 mode='regression',
                 ):
        """
        Preprocessing Class. Cleans a dataset into a workable format.
        Deals with Outliers, Missing Values, duplicate rows, data types (floats, categorical and
        dates), Not a Numbers, Infinities.

        Parameters
        ----------
        target str: Column name of target variable
        num_cols list: Numerical columns, all parsed to integers and floats
        date_cols list: Date columns, all parsed to pd.datetime format
        cat_cols list: Categorical Columns. Currently all one-hot encoded.
        missing_values str: How to deal with missing values ('remove', 'interpolate' or 'mean')
        outlier_removal str: How to deal with outliers ('clip', 'boxplot', 'z-score' or 'none')
        z_score_threshold int: If outlierRemoval='z-score', the threshold is adaptable, default=4.
        folder str: Directory for storing the output files
        version int: Versioning the output files
        mode str: classification / regression
        """
        # Tests
        assert isinstance(target, str)
        assert isinstance(num_cols, list) or num_cols is None
        assert isinstance(date_cols, list) or date_cols is None
        assert isinstance(cat_cols, list) or cat_cols is None
        assert isinstance(missing_values, str)
        mis_values_algo = ['remove_rows', 'remove_cols', 'interpolate', 'mean', 'zero']
        assert missing_values in mis_values_algo, \
            'Missing values algorithm not implemented, pick from {}'.format(', '.join(mis_values_algo))
        out_rem_algo = ['boxplot', 'z-score', 'clip', 'none']
        assert outlier_removal in out_rem_algo, \
            'Outlier Removal algorithm not implemented, pick from {}'.format(', '.join(out_rem_algo))
        assert isinstance(z_score_threshold, int)
        assert isinstance(folder, str)
        assert isinstance(version, int)
        assert mode in ['classification', 'regression'], 'Mode needs to be classification or regression'

        # Parameters
        self.folder = folder if len(folder) == 0 or folder[-1] == '/' else folder + '/'
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.version = version
        self.target = re.sub("[^a-z0-9]", '_', target.lower())
        self.mode = mode
        self.num_cols = [] if num_cols is None else [re.sub('[^a-z0-9]', '_', nc.lower()) for nc in num_cols]
        self.cat_cols = [] if cat_cols is None else [re.sub('[^a-z0-9]', '_', cc.lower()) for cc in cat_cols]
        self.date_cols = [] if date_cols is None else [re.sub('[^a-z0-9]', '_', dc.lower()) for dc in date_cols]
        if self.target in self.num_cols:
            self.num_cols.remove(self.target)

        # Variables
        self.scaler = None
        self.oScaler = None

        # Algorithms
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold

        # Info for Documenting
        self.removedDuplicateRows = 0
        self.removedDuplicateColumns = 0
        self.removedOutliers = 0
        self.imputedMissingValues = 0
        self.removedConstantColumns = 0

    def clean(self, data):
        print('[Data] Data Cleaning Started, ({} x {}) samples'.format(len(data), len(data.keys())))
        if len(data[self.target].unique()) == 2:
            self.mode = 'classification'

        # Note down
        rows, columns = len(data), len(data.keys())

        # Clean Keys
        data = clean_keys(data)

        # Remove Duplicates
        data = self.remove_duplicates(data)
        # Note
        self.removedDuplicateColumns = len(data.keys()) - columns
        self.removedDuplicateRows = len(data) - rows

        # Convert data types
        data = self.convert_data_types(data)

        # Remove outliers
        self.removedOutliers = self.check_outliers(data)
        data = self.remove_outliers(data)
        # Note
        self.imputedMissingValues = np.sum(np.isnan(data.values)) + np.sum(data.values == np.Inf) + \
                                    np.sum(data.values == -np.Inf)

        # Remove missing values
        data = self.remove_missing_values(data)
        # Note
        columns = len(data.keys())

        # Remove Constants
        data = self.remove_constants(data)
        # Note
        self.removedConstantColumns = columns - len(data.keys())

        # Finish
        self._store(data)
        print('[Data] Processing completed, ({} x {}) samples returned'.format(len(data), len(data.keys())))
        return data

    def convert_data_types(self, data):
        # Convert Data Types
        for key in self.date_cols:
            data.loc[:, key] = pd.to_datetime(data[key], errors='coerce', infer_datetime_format=True, utc=True)
        for key in [key for key in data.keys() if key not in self.date_cols and key not in self.cat_cols]:
            data.loc[:, key] = pd.to_numeric(data[key], errors='coerce', downcast='float')
        for key in self.cat_cols:
            if key in data.keys():
                dummies = pd.get_dummies(data[key])
                for dummy_key in dummies.keys():
                    dummies = dummies.rename(
                        columns={dummy_key: key + '_' + re.sub('[^a-z0-9]', '_', str(dummy_key).lower())})
                data = data.drop(key, axis=1).join(dummies)
        return data

    @staticmethod
    def remove_duplicates(data):
        # Remove Duplicates
        data = data.drop_duplicates()
        data = data.loc[:, ~data.columns.duplicated()]
        return data

    @staticmethod
    def remove_constants(data):
        # Remove Constants
        data = data.drop(columns=data.columns[data.nunique() == 1])
        return data

    def check_outliers(self, data):
        if self.outlier_removal == 'boxplot':
            q1 = data.quantile(0.25)
            q3 = data.quantile(0.75)
            return (data > q3).sum().sum() or (data < q1).sum().sum()
        elif self.outlier_removal == 'z-score':
            z_score = (data - data.mean(skipna=True, numeric_only=True)) \
                     / data.std(skipna=True, numeric_only=True)
            return (z_score > self.z_score_threshold).sum().sum()
        elif self.outlier_removal == 'clip':
            return (data > 1e12).sum().sum() + (data < -1e12).sum().sum()

    def remove_outliers(self, data):
        # Remove Outliers
        if self.outlier_removal == 'boxplot':
            q1 = data.quantile(0.25)
            q3 = data.quantile(0.75)
            for key in q1.keys():
                data.loc[data[key] < q1[key] - 1.5 * (q3[key] - q1[key]), key] = np.nan
                data.loc[data[key] > q3[key] + 1.5 * (q3[key] - q1[key]), key] = np.nan
        elif self.outlier_removal == 'z-score':
            z_score = (data - data.mean(skipna=True, numeric_only=True)) \
                     / data.std(skipna=True, numeric_only=True)
            data[z_score > self.z_score_threshold] = np.nan
        elif self.outlier_removal == 'clip':
            data = data.clip(lower=-1e12, upper=1e12)
        return data

    def remove_missing_values(self, data):
        # Remove Missing Values
        data = data.replace([np.inf, -np.inf], np.nan)
        if self.missing_values == 'remove_rows':
            data = data[data.isna().sum(axis=1) == 0]
        elif self.missing_values == 'remove_cols':
            data = data.loc[:, data.isna().sum(axis=0) == 0]
        elif self.missing_values == 'zero':
            data = data.fillna(0)
        elif self.missing_values == 'interpolate':
            ik = np.setdiff1d(data.keys().to_list(), self.date_cols)
            data[ik] = data[ik].interpolate(limit_direction='both')
            if data.isna().sum().sum() != 0:
                data = data.fillna(0)
        elif self.missing_values == 'mean':
            data = data.fillna(data.mean())
        return data

    def _store(self, data):
        # Store cleaned data
        data.to_csv(self.folder + 'Cleaned_v{}.csv'.format(self.version), index_label='index')

    def export_function(self):
        duplicates_code = inspect.getsource(self.remove_duplicates)
        duplicates_code = duplicates_code[duplicates_code.find('\n')+1:]
        function_strings = [
            textwrap.indent(inspect.getsource(clean_keys), '    '),
            duplicates_code.replace('self.', ''),
            inspect.getsource(self.convert_data_types).replace('self.', ''),
            inspect.getsource(self.remove_outliers).replace('self.', ''),
            inspect.getsource(self.remove_missing_values).replace('self.', ''),
        ]
        function_strings = '\n'.join([k[k.find('\n'): k.rfind('\n', 0, k.rfind('\n'))] for k in function_strings])

        return """
        #################
        # Data Cleaning #
        #################
        # Copy vars
        cat_cols, date_cols, target = {}, {}, '{}'
        outlier_removal, missing_values, z_score_threshold = '{}', '{}', '{}'
    """.format(self.cat_cols, self.date_cols, self.target, self.outlier_removal, self.missing_values,
               self.z_score_threshold) + function_strings



import copy
import warnings
import numpy as np
import pandas as pd
import multiprocessing as mp
from datetime import datetime
from scipy.stats import uniform
from scipy.stats import randint
from scipy.stats import loguniform
from sklearn.metrics import SCORERS
from sklearn.model_selection import KFold
from sklearn.experimental import enable_halving_search_cv
from sklearn.model_selection import HalvingRandomSearchCV


class HalvingGridSearch:

    def __init__(self, model, params=None, cv=KFold(n_splits=3), scoring='accuracy',
                 candidates=250, verbose=0):
        # Params
        self.model = model
        self.params = params
        self.cv = cv
        self.scoring = scoring
        self.candidates = candidates
        self.verbose = verbose
        self.samples = None
        self.binary = False

        # Halving settings
        self.resource = 'n_samples'
        self.max_resource = 'auto'
        self.min_resource = 200
        self.set_resources()

        # Tests
        assert model is not None, 'Model not provided'
        assert scoring in SCORERS.keys(), 'Scoring not recognized, pick from sklearn.metrics.SCORERS'
        if params is not None:
            assert isinstance(params, dict), 'Provided params need to be of type dict'
            assert all([hasattr(x, 'rvs') for x in params.values()]), 'Param values should have rvs'

    def set_resources(self):
        if 'CatBoost' in type(self.model).__name__:
            self.resource = 'n_estimators'
            self.max_resource = 3000
            self.min_resource = 250
        if self.model.__module__ == 'sklearn.ensemble._bagging' or \
                self.model.__module__ == 'xgboost.sklearn' or \
                self.model.__module__ == 'lightgbm.sklearn' or \
                self.model.__module__ == 'sklearn.ensemble._forest':
            self.resource = 'n_estimators'
            self.max_resource = 1500
            self.min_resource = 50

    def _get_hyper_params(self):
        # Parameters for both Regression / Classification
        if type(self.model).__name__ == 'LinearRegression':
            return {}
        elif type(self.model).__name__ == 'Lasso' or \
                'Ridge' in type(self.model).__name__:
            return {
                'alpha': uniform(0, 10),
            }
        elif 'SV' in type(self.model).__name__:
            return {
                'gamma': ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1],
                'C': uniform(0, 10),
            }
        elif 'KNeighbors' in type(self.model).__name__:
            return {
                'n_neighbors': randint(5, 50),
                'weights': ['uniform', 'distance'],
                'leaf_size': randint(10, 150),
                'n_jobs': [mp.cpu_count() - 1],
            }

        # Regressor specific hyper parameters
        elif type(self.model).__name__ == 'SGDRegressor':
            return {
                'loss': ['squared_loss', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'],
                'penalty': ['l2', 'l1', 'elasticnet'],
                'alpha': randint(0, 5),
            }
        elif type(self.model).__name__ == 'DecisionTreeRegressor':
            return {
                'criterion': ['mse', 'friedman_mse', 'mae', 'poisson'],
                'max_depth': randint(5, 50),
            }
        elif type(self.model).__name__ == 'AdaBoostRegressor':
            return {
                'n_estimators': randint(25, 250),
                'loss': ['linear', 'square', 'exponential'],
                'learning_rate': uniform(0, 1)
            }
        elif type(self.model).__name__ == 'GradientBoostingRegressor':
            return {
                'loss': ['ls', 'lad', 'huber'],
                'learning_rate': uniform(0, 1),
                'max_depth': randint(3, min(10, int(np.log2(self.samples)))),
                'n_estimators': randint(100, 500),
                'min_samples_leaf': randint(1, min(1000, int(self.samples / 10))),
                'max_features': uniform(0, 1),
                'subsample': uniform(0, 1),
            }
        elif type(self.model).__name__ == 'HistGradientBoostingRegressor':
            return {
                'loss': ['least_squares', 'least_absolute_deviation'],
                'learning_rate': loguniform(0.001, 0.5),
                'max_iter': randint(100, 250),
                'max_leaf_nodes': randint(30, 150),
                'max_depth': randint(3, min(10, int(np.log2(self.samples)))),
                'max_bins': randint(100, 255),
                'l2_regularization': uniform(0, 10),
                'early_stopping': [True],
            }
        elif type(self.model).__name__ == 'RandomForestRegressor':
            return {
                'criterion': ['mse', 'mae'],
                'max_depth': randint(3, min(15, int(np.log2(self.samples)))),
                'max_features': ['auto', 'sqrt'],
                'min_samples_split': randint(2, 50),
                'min_samples_leaf': randint(1, min(1000, int(self.samples / 10))),
                'bootstrap': [True, False],
            }
        elif type(self.model).__name__ == 'CatBoostRegressor':
            return {
                'loss_function': ['MAE', 'RMSE'],
                'learning_rate': loguniform(0.001, 0.5),
                'l2_leaf_reg': uniform(0, 10),
                'depth': randint(3, min(10, int(np.log2(self.samples)))),
                'min_data_in_leaf': randint(1, min(1000, int(self.samples / 10))),
                'max_leaves': randint(10, 250),
            }
        elif type(self.model).__name__ == 'XGBRegressor':
            return {
                'max_depth': randint(3, 15),
                'booster': ['gbtree', 'gblinear', 'dart'],
                'learning_rate': loguniform(0.001, 0.5),
            }
        elif type(self.model).__name__ == 'LGBMRegressor':
            return {
                    'num_leaves': randint(10, 150),
                    'min_child_samples': randint(1, min(1000, int(self.samples / 10))),
                    'min_child_weight': uniform(0, 1),
                    'subsample': uniform(0, 1),
                    'colsample_bytree': uniform(0, 1),
                    'reg_alpha': uniform(0, 1),
                    'reg_lambda': uniform(0, 1),
                }

        # Classification specific hyper parameters
        elif type(self.model).__name__ == 'SGDClassifier':
            return {
                'loss': ['hinge', 'log', 'modified_huber', 'squared_hinge'],
                'penalty': ['l2', 'l1', 'elasticnet'],
                'alpha': uniform(0, 10),
                'max_iter': randint(250, 1000),
            }
        elif type(self.model).__name__ == 'DecisionTreeClassifier':
            return {
                'criterion': ['gini', 'entropy'],
                'max_depth': randint(5, min(50, int(np.log2(self.samples)))),
            }
        elif type(self.model).__name__ == 'AdaBoostClassifier':
            return {
                'n_estimators': randint(25, 250),
                'learning_rate': loguniform(0.001, 0.5)
            }
        elif type(self.model).__name__ == 'BaggingClassifier':
            return {
                # 'n_estimators': [5, 10, 15, 25, 50],
                'max_features': uniform(0, 1),
                'bootstrap': [False, True],
                'bootstrap_features': [True, False],
                'n_jobs': [mp.cpu_count() - 1],
            }
        elif type(self.model).__name__ == 'GradientBoostingClassifier':
            return {
                'loss': ['deviance', 'exponential'],
                'learning_rate': loguniform(0.001, 0.5),
                'max_depth': randint(3, min(15, int(np.log2(self.samples)))),
                'n_estimators': randint(100, 500),
                'min_samples_leaf': randint(1, min(1000, int(self.samples / 10))),
                'max_features': uniform(0, 1),
                'subsample': uniform(0, 1),
            }
        elif type(self.model).__name__ == 'HistGradientBoostingClassifier':
            return {
                'learning_rate': loguniform(0.001, 0.5),
                'max_iter': randint(100, 250),
                'max_leaf_nodes': randint(30, 150),
                'max_depth': randint(3, min(10, int(np.log2(self.samples)))),
                'min_samples_leaf': randint(1, min(1000, int(self.samples / 10))),
                'l2_regularization': loguniform(0.0001, 10),
                'max_bins': randint(100, 255),
                'early_stopping': [True]
            }
        elif type(self.model).__name__ == 'RandomForestClassifier':
            return {
                'criterion': ['gini', 'entropy'],
                'max_depth': randint(3, min(15, int(np.log2(self.samples)))),
                'max_features': ['auto', 'sqrt'],
                'min_samples_split': randint(2, 50),
                'min_samples_leaf': randint(1, min(1000, int(self.samples / 10))),
                'bootstrap': [True, False],
            }
        elif type(self.model).__name__ == 'CatBoostClassifier':
            return {
                'loss_function': ['Logloss' if self.binary else 'MultiClass'],
                'eval_metric': ['Logloss'],
                'learning_rate': loguniform(0.001, 0.5),
                'l2_leaf_reg': loguniform(0.0001, 10),
                'depth': randint(1,  min(10, int(np.log2(self.samples)))),
                'min_data_in_leaf': randint(1, min(1000, int(self.samples / 10))),
                'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
            }
        elif type(self.model).__name__ == 'XGBClassifier':
            return {
                "objective": ["binary:logistic"] if self.binary else ['multi:softprob'],
                'max_depth': randint(3,  min(10, int(np.log2(self.samples)))),
                'booster': ['gbtree', 'gblinear', 'dart'],
                'learning_rate': loguniform(0.001, 0.5),
                'verbosity': [0],
                'scale_pos_weight': uniform(0, 100),
                "alpha": loguniform(0.0001, 1)
            }
        elif type(self.model).__name__ == 'LGBMClassifier':
            return {
                    "objective": "binary" if self.binary else 'multiclass',
                    "verbosity": [-1],
                    "boosting_type": ["gbdt"],
                    "lambda_l1": loguniform(1e-3, 10),
                    "lambda_l2": loguniform(1e-3, 10),
                    'num_leaves': randint(2, 256),
                    'min_child_samples': randint(1, min(1000, int(self.samples / 10))),
                    'min_child_weight': loguniform(1e-5, 1),
                    'subsample': uniform(0, 1),
                    'colsample_bytree': uniform(0, 1),
        }

        else:
            # Raise error if nothing is returned
            warnings.warn('Hyper parameter tuning not implemented for {}'.format(type(self.model).__name__))
            return {}

    def fit(self, x, y):
        # Update minimum resource for samples (based on dataset)
        if self.resource == 'n_samples':
            self.min_resource = int(0.2 * len(x)) if len(x) > 5000 else len(x)

        # Parameters
        if len(np.unique(y)) == 2:
            self.binary = True
        self.samples = len(y)
        if self.params is None:
            self.params = self._get_hyper_params()

        # Set up and run grid search
        halving_random_search = HalvingRandomSearchCV(self.model,
                                                      self.params,
                                                      n_candidates=self.candidates,
                                                      resource=self.resource,
                                                      max_resources=self.max_resource,
                                                      min_resources=self.min_resource,
                                                      cv=self.cv,
                                                      scoring=self.scoring,
                                                      factor=3,
                                                      n_jobs=mp.cpu_count() - 1,
                                                      verbose=self.verbose)
        halving_random_search.fit(x, y)

        # Parse results
        scikit_results = pd.DataFrame(halving_random_search.cv_results_)
        results = pd.DataFrame({
            'date': datetime.today().strftime('%d %b %y'),
            'model': type(self.model).__name__,
            'params': scikit_results['params'],
            'mean_objective': scikit_results['mean_test_score'],
            'std_objective': scikit_results['std_test_score'],
            'worst_case': scikit_results['mean_test_score'] - scikit_results['std_test_score'],
            'mean_time': scikit_results['mean_fit_time'],
            'std_time': scikit_results['std_fit_time'],
        })

        # Update resource in results
        if self.resource != 'n_samples':
            for i in range(len(results)):
                results.loc[results.index[i], 'params'][self.resource] = self.max_resource

        return results



import unittest
import pandas as pd
from sklearn.datasets import load_iris
from sklearn.datasets import load_boston
from Amplo.AutoML import DataProcessing
from Amplo.Utils import check_dataframe_quality


class TestDataProcessing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.classification, y = load_iris(return_X_y=True, as_frame=True)
        cls.classification['target'] = y
        x, y = load_boston(return_X_y=True)
        cls.regression = pd.DataFrame({'target': y})
        for i in range(x.shape[1]):
            cls.regression['feature_{}'.format(i)] = x[:, i]

    def test_regression(self):
        dp = DataProcessing('target', mode='regression', folder='tmp/')
        cleaned = dp.clean(self.regression)
        assert check_dataframe_quality(cleaned)

    def test_classification(self):
        dp = DataProcessing('target', mode='classification', folder='tmp/')
        cleaned = dp.clean(self.classification)
        assert check_dataframe_quality(cleaned)
